"""Mock Kubernetes API for testing."""

from __future__ import annotations

import copy
import datetime
import json
import os
import re
from collections import defaultdict
from collections.abc import AsyncGenerator, Awaitable, Callable, Iterator
from datetime import timedelta
from typing import Any, Protocol
from unittest.mock import AsyncMock, Mock, patch

try:
    from kubernetes_asyncio import client, config
    from kubernetes_asyncio.client import (
        ApiException,
        CoreV1Event,
        CoreV1EventList,
        V1ConfigMap,
        V1DeleteOptions,
        V1Ingress,
        V1IngressList,
        V1IngressStatus,
        V1Job,
        V1JobList,
        V1JobStatus,
        V1LoadBalancerStatus,
        V1Namespace,
        V1NamespaceList,
        V1NetworkPolicy,
        V1Node,
        V1NodeList,
        V1ObjectMeta,
        V1ObjectReference,
        V1PersistentVolume,
        V1PersistentVolumeClaim,
        V1PersistentVolumeClaimList,
        V1PersistentVolumeList,
        V1Pod,
        V1PodList,
        V1PodStatus,
        V1ResourceQuota,
        V1Secret,
        V1Service,
        V1ServiceAccount,
        V1ServiceAccountList,
        V1ServiceList,
        V1Status,
    )
except ImportError as e:
    raise ImportError(
        "The safir.testing.kubernetes module requires the kubernetes extra. "
        "Install it with `pip install safir[kubernetes]`."
    ) from e

from safir.asyncio import AsyncMultiQueue
from safir.datetime import current_datetime

__all__ = [
    "MockKubernetesApi",
    "patch_kubernetes",
]


def _parse_label_selector(label_selector: str) -> dict[str, str]:
    """Parse a label selector.

    Only equality is supported (with ``=`` or ``==``).

    Parameters
    ----------
    label_selector
        Label selector string to parse.

    Returns
    -------
    dict of str
        Dictionary of required labels to their required values.

    Raises
    ------
    AssertionError
        Raised if the label selector string is not a supported syntax.
    """
    result = {}
    for requirement in label_selector.split(","):
        match = re.match(r"([^!=]+)==?(.*)", requirement)
        assert match
        assert match.group(1)
        assert match.group(2)
        result[match.group(1)] = match.group(2)
    return result


def _check_labels(
    obj_labels: dict[str, str] | None, label_selector: str | None
) -> bool:
    """Check whether an object's labels match a label selector.

    Parameters
    ----------
    obj_labels
        Kubernetes object labels.
    label_selector
        Label selector in string form.

    Returns
    -------
    bool
        Whether this object matches the label selector.
    """
    # Everything matches the absence of a selector.
    if not label_selector:
        return True

    # If there are no labels but a non-empty selector, it doesn't match.
    if not obj_labels:
        return False

    # Check that all labels match the labels of the object.
    labels = _parse_label_selector(label_selector)
    for label in labels:
        if label not in obj_labels or labels[label] != obj_labels[label]:
            return False
    return True


class _KubernetesModel(Protocol):
    """Protocol describing common features of Kubernetes objects.

    The kubernetes_asyncio_ library doesn't provide full typing of its methods
    or objects, so use a protocol to describe the functionality that we rely
    on when passing a generic object around.
    """

    def to_dict(
        self,
        serialize: bool = False,  # noqa: FBT001, FBT002
    ) -> dict[str, Any]: ...


class _DatetimeSerializer(json.JSONEncoder):
    """Allows serialization of `datetime.datetime` objects.

    If one is encountered, serializes it with the object's isoformat() method.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        return super().default(o)


class _EventStream:
    """Holds the data for a stream of watchable events.

    This is an internal implementation detail of the Kubernetes mock. It holds
    a stream of watchable events and a list of `asyncio.Event` triggers. A
    watch can register interest in this event stream, in which case its
    trigger will be notified when anything new is added to the event stream.

    The events are generic dicts, which will be interpreted by the Kubernetes
    library differently depending on which underlying API is using this data
    structure.
    """

    def __init__(self) -> None:
        self._queue = AsyncMultiQueue[dict[str, Any]]()

    @property
    def next_resource_version(self) -> str:
        """Resource version of the next event.

        This starts with ``1`` to ensure that a resource version of ``0`` is
        special and means to return all known events, so it must be adjusted
        when indexing into a list of events.
        """
        return str(self._queue.qsize() + 1)

    def add_custom_event(self, action: str, obj: dict[str, Any]) -> None:
        """Add a new event for a custom object and notify all watchers.

        Parameters
        ----------
        action
            Action of the event.
        obj
            Custom object triggering the event.
        """
        self._queue.put({"type": action, "object": obj})

    def add_event(self, action: str, obj: _KubernetesModel) -> None:
        """Add a new event and notify all watchers.

        Parameters
        ----------
        action
            Action of the event.
        obj
            Object triggering the event.
        """
        event = {"type": action, "object": obj.to_dict(serialize=True)}
        self._queue.put(event)

    def build_watch_response(
        self,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
    ) -> Mock:
        """Construct a response to a watch request.

        Wraps a watch iterator in the appropriate mock to behave as expected
        when called from ``kubernetes_asyncio``. This simulates an aiohttp
        response.

        Parameters
        ----------
        resource_version
            Starting resource version for the watch. The resource versions
            must be the position in the events stream, stringified.
        timeout_seconds
            How long to watch. This is total elapsed time, not the time
            between events. It matches the corresponding parameter in the
            Kubernetes API. If `None`, watch forever.
        field_selector
            Limit the returned events to the objects matching the selector.
            If set, it must be set to ``metadata.name=...`` to match a
            specific object name.
        label_selector
            Limit the returned events to the objects matching this label
            selector.

        Returns
        -------
        Mock
            Simulation of an aiohttp response whose ``readline`` method wraps
            the iterator.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        """
        generator = self._build_watcher(
            resource_version, timeout_seconds, field_selector, label_selector
        )

        async def readline() -> bytes:
            return await generator.__anext__()

        response = Mock()
        response.content.readline = AsyncMock()
        response.content.readline.side_effect = readline
        return response

    def _build_watcher(
        self,
        resource_version: str | None,
        timeout_seconds: int | None,
        field_selector: str | None,
        label_selector: str | None,
    ) -> AsyncGenerator[bytes]:
        """Construct a watcher for this event stream.

        Parameters
        ----------
        resource_version
            Starting resource version for the watch. The resource version
            must be the position in the events stream, stringified. If not
            given, start after the most recent event.
        timeout_seconds
            How long to watch. This is total elapsed time, not the time
            between events. It matches the corresponding parameter in the
            Kubernetes API. If `None`, watch forever.
        field_selector
            Limit the returned events to the objects matching the selector.
            If set, it must be set to ``metadata.name=...`` to match a
            specific object name.
        label_selector
            Limit the returned events to the objects matching this label
            selector.

        Returns
        -------
        AsyncGenerator
            An async iterator that will return each new event in the stream,
            encoded as expected by the Kubernetes API, until the timeout
            occurs.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        """
        timeout = None
        if timeout_seconds is not None:
            timeout = timedelta(seconds=timeout_seconds)

        # Parse the field selector, if one was provided.
        name = None
        if field_selector:
            match = re.match(r"metadata\.name=(.*)$", field_selector)
            assert match
            assert match.group(1)
            name = match.group(1)

        # Construct the iterator.
        async def next_event() -> AsyncGenerator[bytes]:
            if resource_version:
                start = int(resource_version)
            else:
                start = self._queue.qsize()
            try:
                async for event in self._queue.aiter_from(start, timeout):
                    if name and event["object"]["metadata"]["name"] != name:
                        continue
                    labels = event["object"]["metadata"]["labels"]
                    if not _check_labels(labels, label_selector):
                        continue
                    yield json.dumps(event, cls=_DatetimeSerializer).encode()
            except TimeoutError:
                yield b""

        # Return the iterator.
        return next_event()


class MockKubernetesApi:
    """Mock Kubernetes API for testing.

    This object simulates (with almost everything left out) the ``BatchV1Api``,
    ``CoreV1Api``, ``CustomObjectApi``, and ``NetworkingV1Api`` client
    objects while keeping simple internal state. It is intended to be used
    as a mock inside tests.

    Methods ending with ``_for_test`` are outside of the API and are intended
    for use by the test suite.

    This mock does not enforce namespace creation before creating objects in a
    namespace. Creating an object in a namespace will implicitly create that
    namespace if it doesn't exist. However, it will not store a
    ``V1Namespace`` object or emit events for `list_namespace` watches. To
    verify that a namespace was properly created (although not the order of
    creation), check for the namespace object explicitly with `read_namespace`
    or `list_namespace`.

    Objects stored with ``create_*`` or ``replace_*`` methods are **NOT**
    copied. The object provided will be stored, so changing that object will
    change the object returned by subsequent API calls. Likewise, the object
    returned by ``read_*`` calls will be the same object stored in the mock,
    and changing it will change the mock's data. (Sometimes this is the
    desired behavior, sometimes it isn't; we had to pick one and this is the
    approach we picked.)

    Not all methods are implemented for each kind, and not all ``list_*`` APIs
    support watches. In general, APIs are only implemented when some other
    package needs them.

    Attributes
    ----------
    initial_pod_phase
        String value to set the status of pods to when created. If this is set
        to ``Running`` (the default), a pod start event will also be
        generated when the pod is created.
    error_callback
        If set, called with the method name and any arguments whenever any
        Kubernetes API method is called and before it takes any acttion. This
        can be used for fault injection for testing purposes.

    Notes
    -----
    This class is normally not instantiated directly. Instead, call the
    `patch_kubernetes` function from a fixture to set up the mock. This is
    also why it is configurable by setting attributes rather than constructor
    arguments; the individual test usually doesn't have control of the
    constructor.
    """

    def __init__(self) -> None:
        self.error_callback: Callable[..., None] | None = None
        self.initial_pod_phase = "Running"

        self._cluster_objects: dict[str, dict[str, Any]] = {}
        self._create_hooks: dict[str, Callable[..., Awaitable[None]]] = {}
        self._custom_kinds: dict[str, str] = {}
        self._nodes = V1NodeList(items=[])
        self._objects: dict[str, dict[str, dict[str, Any]]] = {}
        self._events: defaultdict[str, list[CoreV1Event]] = defaultdict(list)
        self._namespace_stream = _EventStream()
        self._event_streams: defaultdict[str, defaultdict[str, _EventStream]]
        self._event_streams = defaultdict(lambda: defaultdict(_EventStream))
        self._cluster_event_streams: defaultdict[str, _EventStream]
        self._cluster_event_streams = defaultdict(_EventStream)

    def get_all_objects_for_test(self, kind: str) -> list[Any]:
        """Return all objects of a given kind sorted by namespace and name.

        Parameters
        ----------
        kind
            The Kubernetes kind, such as ``Secret`` or ``Pod``. This is
            case-sensitive.

        Returns
        -------
        list of Any
            All objects of that kind found in the mock, sorted by namespace
            and then name.
        """
        key = self._custom_kinds.get(kind, kind)
        results = []
        for namespace in sorted(self._objects.keys()):
            if key not in self._objects[namespace]:
                continue
            for _name, obj in sorted(self._objects[namespace][key].items()):
                results.append(obj)
        return results

    def get_namespace_objects_for_test(self, namespace: str) -> list[Any]:
        """Return all objects in the given namespace.

        Parameters
        ----------
        namespace
            Name of the namespace.

        Returns
        -------
        list of Any
            All objects found in that namespace, sorted by kind and then name.
            Due to how objects are stored in the mock, we can't distinguish
            between a missing namespace and a namespace with no objects. In
            both cases, the empty list is returned.
        """
        if namespace not in self._objects:
            return []
        result = []
        for kind in sorted(self._objects[namespace].keys()):
            for _, body in sorted(self._objects[namespace][kind].items()):
                result.append(body)
        return result

    def register_create_hook_for_test(
        self, kind: str, hook: Callable[..., Awaitable[None]] | None
    ) -> None:
        """Register a creation hook for a particular kind of object.

        The registered hook replaces any existing hook for that object.

        Parameters
        ----------
        kind
            Kind of object. This must use the same capitalization convention
            as Kubernetes itself does (for example, ``ServiceAccount`, not
            ``serviceaccount``).
        hook:
            Async callback that is called after a new object of this type is
            created, or `None` to remove any hooks for that kind of object.
            The hook will be called with one argument, the newly created
            object.
        """
        if hook:
            self._create_hooks[kind] = hook
        elif kind in self._create_hooks:
            del self._create_hooks[kind]

    def set_nodes_for_test(self, nodes: list[V1Node]) -> None:
        """Set the node structures that will be returned by `list_node`.

        Parameters
        ----------
        nodes
            New node list to return.
        """
        self._nodes = nodes

    # CUSTOM OBJECT API

    async def create_namespaced_custom_object(
        self,
        group: str,
        version: str,
        namespace: str,
        plural: str,
        body: dict[str, Any],
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a new custom namespaced object.

        Parameters
        ----------
        group
            API group for this custom object.
        version
            API version for this custom object.
        namespace
            Namespace in which to create the object.
        plural
            API plural for this custom object.
        body
            Custom object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error(
            "create_namespaced_custom_object",
            group,
            version,
            namespace,
            plural,
            body,
        )
        api_version = body.get("api_version", body["apiVersion"])
        assert api_version == f"{group}/{version}"
        key = f"{group}/{version}/{plural}"
        if body["kind"] in self._custom_kinds:
            assert key == self._custom_kinds[body["kind"]]
        else:
            self._custom_kinds[body["kind"]] = key
        if body["metadata"].get("namespace"):
            assert namespace == body["metadata"]["namespace"]
        else:
            body["metadata"]["namespace"] = namespace
        await self._store_object(
            namespace, key, body["metadata"]["name"], body
        )

    async def delete_namespaced_custom_object(
        self,
        group: str,
        version: str,
        namespace: str,
        plural: str,
        name: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> Any:
        """Delete a custom namespaced object.

        Parameters
        ----------
        group
            API group for this custom object.
        version
            API version for this custom object.
        namespace
            Namespace in which to delete the object.
        plural
            API plural for this custom object.
        name
            Custom object to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status if object was deleted.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("delete_namespaced_custom_object", name, namespace)
        key = f"{group}/{version}/{plural}"
        return self._delete_object(namespace, key, name, propagation_policy)

    async def get_namespaced_custom_object(
        self,
        group: str,
        version: str,
        namespace: str,
        plural: str,
        name: str,
        *,
        _request_timeout: float | None = None,
    ) -> dict[str, Any]:
        """Retrieve a namespaced custom object.

        Parameters
        ----------
        group
            API group for this custom object.
        version
            API version for this custom object.
        namespace
            Namespace in which to create the object.
        plural
            API plural for this custom object.
        name
            Name of the object to retrieve.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        dict of Any
            Body of the custom object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the object does not exist.
        """
        self._maybe_error(
            "get_namespaced_custom_object",
            group,
            version,
            namespace,
            plural,
            name,
        )
        return self._get_object(namespace, f"{group}/{version}/{plural}", name)

    async def list_cluster_custom_object(
        self,
        group: str,
        version: str,
        plural: str,
        *,
        _request_timeout: float | None = None,
    ) -> dict[str, list[dict[str, Any]]]:
        """List all custom objects in the cluster.

        Parameters
        ----------
        group
            API group for this custom object.
        version
            API version for this custom object.
        plural
            API plural for this custom object.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        dict
            Dictionary with one key, ``items``, whose value is a list of all
            the specified custom objects stored in the cluster.
        """
        self._maybe_error("list_cluster_custom_object", group, version, plural)
        key = f"{group}/{version}/{plural}"
        results: list[dict[str, Any]] = []
        for namespace in self._objects:
            results.extend(self._objects[namespace].get(key, {}).values())
        return {"items": results}

    async def list_namespaced_custom_object(
        self,
        group: str,
        version: str,
        namespace: str,
        plural: str,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1IngressList | Mock:
        """List custom objects in a namespace.

        This does support watches.

        Parameters
        ----------
        group
            API group for this custom object.
        version
            API version for this custom object.
        namespace
            Namespace of custom object to list.
        plural
            API plural for this custom object.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            custom object name and only custom objects matching that name will
            be returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        dict or unittest.mock.Mock
            List of custom objects in that namespace as a dict with one key,
           ``items``, when not called as a watch. If called as a watch,
           returns a mock ``aiohttp.Response`` with a ``readline`` metehod
           that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error(
            "list_namespaced_custom_object",
            group,
            version,
            namespace,
            plural,
            field_selector,
        )
        key = f"{group}/{version}/{plural}"
        if namespace not in self._objects:
            msg = f"Namespace {namespace} not found"
            raise ApiException(status=404, reason=msg)
        if not watch:
            objs = self._list_objects(
                namespace, key, field_selector, label_selector
            )
            return {"items": objs}

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace][key]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def patch_namespaced_custom_object_status(
        self,
        group: str,
        version: str,
        namespace: str,
        plural: str,
        name: str,
        body: list[dict[str, Any]],
        *,
        _request_timeout: float | None = None,
    ) -> dict[str, Any]:
        """Patch the status of a namespaced custom object.

        Parameters
        ----------
        group
            API group for this custom object.
        version
            API version for this custom object.
        namespace
            Namespace in which to create the object.
        plural
            API plural for this custom object.
        name
            Name of the object to retrieve.
        body
            Body of the patch. The only patch supported is one with ``op`` of
            ``replace`` and ``path`` of ``/status``.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        dict of Any
            Modified body of the custom object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the object does not exist.
        AssertionError
            Raised if any other type of patch is provided.
        """
        self._maybe_error(
            "patch_namespaced_custom_object_status",
            group,
            version,
            namespace,
            plural,
            body,
        )
        key = f"{group}/{version}/{plural}"
        obj = copy.deepcopy(self._get_object(namespace, key, name))
        for change in body:
            assert change["op"] == "replace"
            assert change["path"] == "/status"
            obj["status"] = change["value"]
        await self._store_object(namespace, key, name, obj, replace=True)
        return obj

    async def replace_namespaced_custom_object(
        self,
        group: str,
        version: str,
        namespace: str,
        plural: str,
        name: str,
        body: dict[str, Any],
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Replace a custom namespaced object.

        Parameters
        ----------
        group
            API group for this custom object.
        version
            API version for this custom object.
        namespace
            Namespace in which to create the object.
        plural
            API plural for this custom object.
        body
            New contents of custom object.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 if the object does not exist.
        """
        self._maybe_error(
            "replace_namespaced_custom_object",
            group,
            version,
            namespace,
            plural,
            name,
            body,
        )
        api_version = body.get("api_version", body["apiVersion"])
        assert api_version == f"{group}/{version}"
        key = f"{group}/{version}/{plural}"
        assert key == self._custom_kinds[body["kind"]]
        assert namespace == body["metadata"]["namespace"]
        await self._store_object(namespace, key, name, body, replace=True)

    # CONFIGMAP API

    async def create_namespaced_config_map(
        self,
        namespace: str,
        body: V1ConfigMap,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a ``ConfigMap`` object.

        Parameters
        ----------
        namespace
            Namespace in which to store the object.
        body
            Object to store.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_config_map", namespace, body)
        self._update_metadata(body, "v1", "ConfigMap", namespace)
        name = body.metadata.name
        await self._store_object(namespace, "ConfigMap", name, body)

    async def delete_namespaced_config_map(
        self,
        name: str,
        namespace: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete a ``ConfigMap`` object.

        Parameters
        ----------
        name
            Name of object.
        namespace
            Namespace of object.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status if object was deleted.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the object does not exist.
        """
        self._maybe_error("delete_namespaced_config_map", name, namespace)
        return self._delete_object(
            namespace, "ConfigMap", name, propagation_policy
        )

    async def read_namespaced_config_map(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1ConfigMap:
        """Read a config map object.

        Parameters
        ----------
        name
            Name of object.
        namespace
            Namespace of object.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1ConfigMap
            Corresponding object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the config map was not found.
        """
        self._maybe_error("read_namespaced_config_map", name, namespace)
        return self._get_object(namespace, "ConfigMap", name)

    # EVENTS API

    async def create_namespaced_event(
        self,
        namespace: str,
        body: CoreV1Event,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Store a new namespaced event.

        This uses the old core event API, not the new Events API.

        Parameters
        ----------
        namespace
            Namespace of the event.
        body
            New event to store.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.
        """
        self._maybe_error("create_namespaced_event", namespace, body)
        self._update_metadata(body, "v1", "Event", namespace)
        stream = self._event_streams[namespace]["Event"]
        body.metadata.resource_version = stream.next_resource_version
        stream.add_event("ADDED", body)
        self._events[namespace].append(body)
        if "Event" in self._create_hooks:
            hook = self._create_hooks["Event"]
            await hook(body)

    async def list_namespaced_event(
        self,
        namespace: str,
        *,
        field_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> CoreV1EventList | Mock:
        """List namespaced events.

        This uses the old core event API, not the new Events API. It does
        support watches.

        Parameters
        ----------
        namespace
            Namespace to watch for events.
        field_selector
            Which events to retrieve when performing a watch. Currently, this
            is ignored.
        resource_version
            Where to start in the event stream when performing a watch.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.CoreV1EventList or unittest.mock.Mock
            List of events, when not called as a watch. If called as a watch,
            returns a mock ``aiohttp.Response`` with a ``readline`` method
            that yields the events.
        """
        self._maybe_error("list_namespaced_event", namespace)
        if not watch:
            return CoreV1EventList(items=self._events[namespace])

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace]["Event"]
        return stream.build_watch_response(resource_version, timeout_seconds)

    # INGRESS API

    async def create_namespaced_ingress(
        self,
        namespace: str,
        body: V1Ingress,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create an ingress object.

        In real life, it usually takes some time for the Ingress controller
        to set the ``status.load_balancer.ingress`` field.  It appears that
        status.load_balancer is created when the Ingress is but is empty,
        so that's what we do too.

        Use ``patch_namespaced_ingress_status()`` to update the field to
        indicate that the ingress is ready.

        Parameters
        ----------
        namespace
            Namespace in which to create the object.
        body
            Object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_ingress", namespace, body)
        self._update_metadata(
            body, "networking.k8s.io/v1", "Ingress", namespace
        )
        body.status = V1IngressStatus(
            load_balancer=V1LoadBalancerStatus(ingress=[])
        )
        name = body.metadata.name
        await self._store_object(namespace, "Ingress", name, body)

    async def delete_namespaced_ingress(
        self,
        name: str,
        namespace: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete an ingress object.

        Parameters
        ----------
        name
            Name of ingress to delete.
        namespace
            Namespace of ingress to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the ingress was not found.
        """
        self._maybe_error("delete_namespaced_ingress", name, namespace)
        return self._delete_object(
            namespace, "Ingress", name, propagation_policy
        )

    async def list_namespaced_ingress(
        self,
        namespace: str,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1IngressList | Mock:
        """List ingress objects in a namespace.

        This does support watches.

        Parameters
        ----------
        namespace
            Namespace of ingresses to list.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            ingress name and only ingresses matching that name will be
            returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1IngressList or unittest.mock.Mock
            List of ingresses in that namespace, when not called as a watch. If
            called as a watch, returns a mock ``aiohttp.Response`` with a
            ``readline`` metehod that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error("list_namespaced_ingress", namespace, field_selector)
        if namespace not in self._objects:
            msg = f"Namespace {namespace} not found"
            raise ApiException(status=404, reason=msg)
        if not watch:
            ingresses = self._list_objects(
                namespace, "Ingress", field_selector, label_selector
            )
            return V1IngressList(kind="Ingress", items=ingresses)

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace]["Ingress"]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def patch_namespaced_ingress_status(
        self,
        name: str,
        namespace: str,
        body: list[dict[str, Any]],
        *,
        _request_timeout: float | None = None,
    ) -> V1Secret:
        """Patch the status of an ingress object.

        Parameters
        ----------
        name
            Name of ingress object.
        namespace
            Namespace of ingress object.
        body
            Patches to apply. Only patches with ``op`` of ``replace`` are
            supported, and only with ``path`` of
            ``/status/loadBalancer/ingress``.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Ingress
            Corresponding object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the ingress does not exist.
        AssertionError
            Raised if any other type of patch was provided.
        """
        self._maybe_error("patch_namespaced_ingress", name, namespace)
        ingress = copy.deepcopy(self._get_object(namespace, "Ingress", name))
        for change in body:
            assert change["op"] == "replace"
            assert change["path"] == "/status/loadBalancer/ingress"
            ingress.status.load_balancer.ingress = change["value"]
        await self._store_object(
            namespace, "Ingress", name, ingress, replace=True
        )
        return ingress

    async def read_namespaced_ingress(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Ingress:
        """Read a ingress object.

        Parameters
        ----------
        name
            Name of the ingress.
        namespace
            Namespace of the ingress.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Ingress
            Ingress object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the ingress was not found.
        """
        self._maybe_error("read_namespaced_ingress", name, namespace)
        return self._get_object(namespace, "Ingress", name)

    # JOB API

    async def create_namespaced_job(
        self,
        namespace: str,
        body: V1Job,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a job object.

        A pod corresponding to this job will also be created. The pod will
        have a label ``job-name`` set to the name of the ``Job`` object, and
        will honor the ``name`` or ``generateName`` metadata from the pod spec
        in the ``Job``. If neither is set, it will use the job name followed
        by ``-`` as the base for a generated name.

        If ``initial_pod_phase`` on the mock is set to ``Running``, the
        ``status.active`` field of the job will be set to 1.

        Parameters
        ----------
        namespace
            Namespace in which to create the object.
        body
            Object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_job", namespace, body)
        self._update_metadata(body, "batch/v1", "Job", namespace)
        await self._store_object(namespace, "Job", body.metadata.name, body)

        # Normally, Kubernetes will immediately spawn a Pod using the
        # specification in the Job. Simulate that here. We have to copy the
        # components of the spec metadata so that we don't modify the Job when
        # we flesh out the metadata of the Pod.
        name = body.metadata.name
        metadata = V1ObjectMeta()
        if body.spec.template.metadata:
            source = body.spec.template.metadata
            metadata.name = source.name
            metadata.generate_name = source.generate_name
            if source.labels:
                metadata.labels = source.labels.copy()
            if source.annotations:
                metadata.annotations = source.annotations.copy()
        pod = V1Pod(metadata=metadata, spec=body.spec.template.spec)
        if not pod.metadata.name:
            if not pod.metadata.generate_name:
                pod.metadata.generate_name = f"{name}-"
            base = pod.metadata.generate_name
            pod.metadata.name = base + os.urandom(3).hex()[:5]
        pod.metadata.labels["job-name"] = name
        pod.metadata.namespace = namespace
        await self.create_namespaced_pod(namespace, pod)
        if pod.status.phase == "Running":
            body.status = V1JobStatus(active=1)

    async def delete_namespaced_job(
        self,
        name: str,
        namespace: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete a job object.

        Will also propagate to pods.

        Parameters
        ----------
        name
            Name of job to delete.
        namespace
            Namespace of job to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the job was not found.
        """
        self._maybe_error("delete_namespaced_job", name, namespace)

        # This simulates a foreground deletion, where the Job is blocked
        # from deletion until all its pods are deleted. We also use it for
        # background deletion for the time being, since it should be close
        # enough.
        pods = await self.list_namespaced_pod(
            namespace, label_selector=f"job-name=={name}"
        )
        if propagation_policy != "Orphan":
            for pod in pods.items:
                await self.delete_namespaced_pod(pod.metadata.name, namespace)

        # Perform the actual deletion.
        return self._delete_object(namespace, "Job", name, propagation_policy)

    async def list_namespaced_job(
        self,
        namespace: str,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1JobList | Mock:
        """List job objects in a namespace.

        This does support watches.

        Parameters
        ----------
        namespace
            Namespace of jobs to list.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            job name and only jobs matching that name will be returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1JobList or unittest.mock.Mock
            List of jobs in that namespace, when not called as a watch. If
            called as a watch, returns a mock ``aiohttp.Response`` with a
            ``readline`` metehod that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error("list_namespaced_job", namespace, field_selector)
        if namespace not in self._objects:
            msg = f"Namespace {namespace} not found"
            raise ApiException(status=404, reason=msg)
        if not watch:
            jobs = self._list_objects(
                namespace, "Job", field_selector, label_selector
            )
            return V1PodList(kind="Job", items=jobs)

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace]["Job"]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def read_namespaced_job(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Job:
        """Read a job object.

        Parameters
        ----------
        name
            Name of the job.
        namespace
            Namespace of the job.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Job
            Job object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the job was not found.
        """
        self._maybe_error("read_namespaced_job", name, namespace)
        return self._get_object(namespace, "Job", name)

    # NAMESPACE API

    async def create_namespace(
        self,
        body: V1Namespace,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a namespace.

        The mock doesn't truly track namespaces since it autocreates them when
        an object is created in that namespace (maybe that behavior should be
        optional). However, this method detects conflicts and stores the
        ``V1Namespace`` object so that it can be verified.

        Parameters
        ----------
        body
            Namespace to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the namespace already exists.
        """
        self._maybe_error("create_namespace", body)
        self._update_metadata(body, "v1", "Namespace")
        name = body.metadata.name
        if name in self._objects:
            msg = f"Namespace {name} already exists"
            raise ApiException(status=409, reason=msg)
        await self._store_object(name, "Namespace", name, body)

    async def delete_namespace(
        self,
        name: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> None:
        """Delete a namespace.

        This also immediately removes all objects in the namespace.

        Parameters
        ----------
        name
            Namespace to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error("delete_namespace")
        if name not in self._objects:
            raise ApiException(status=404, reason=f"{name} not found")
        try:
            body = await self.read_namespace(name)
            self._namespace_stream.add_event("DELETED", body)
        except ApiException:
            pass
        del self._objects[name]

    async def read_namespace(
        self,
        name: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Namespace:
        """Return the namespace object for a namespace.

        This will only return a namespace object if one was created via
        `create_namespace`, not if one was implicitly added by creating some
        other object.

        Parameters
        ----------
        name
            Name of namespace to retrieve.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Namespace
            Corresponding namespace object. If `create_namespace` has not been
            called, but the namespace was added implicitly, an exception will
            be raised.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error("read_namespace", name)
        return self._get_object(name, "Namespace", name)

    async def list_namespace(
        self,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1NamespaceList | Mock:
        """List namespaces.

        This does support watches. Only namespaces that are explicitly
        created with `create_namespace` will be shown, not ones that were
        implicitly created by creating some other object.

        Parameters
        ----------
        namespace
            Namespace of jobs to list.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            namespace name and only the namespace matching that name will be
            returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1JobList or unittest.mock.Mock
            List of namespaces, when not called as a watch. If called as a
            watch, returns a mock ``aiohttp.Response`` with a ``readline``
            metehod that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error("list_namespace")

        # We annoyingly have to duplicate a bunch of logic from _list_objects
        # because namespaces aren't stored under a single namespace.
        if not watch:
            if field_selector:
                match = re.match(r"metadata\.name=(.*)$", field_selector)
                if not match or not match.group(1):
                    msg = f"Field selector {field_selector} not supported"
                    raise ValueError(msg)
                try:
                    name = match.group(1)
                    obj = self._get_object(name, "Namespace", name)
                    if _check_labels(obj.metadata.labels, label_selector):
                        return [obj]
                    else:
                        return []
                except ApiException:
                    return []
            namespaces = []
            for name in self._objects:
                try:
                    namespace = await self.read_namespace(name)
                except ApiException:
                    continue
                if _check_labels(namespace.metadata.labels, label_selector):
                    namespaces.append(namespace)
            return V1NamespaceList(items=namespaces)

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        return self._namespace_stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    # NETWORKPOLICY API

    async def create_namespaced_network_policy(
        self,
        namespace: str,
        body: V1NetworkPolicy,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a network policy object.

        Parameters
        ----------
        namespace
            Namespace in which to create the object.
        body
            Object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_network_policy", namespace, body)
        self._update_metadata(
            body, "networking.k8s.io/v1", "NetworkPolicy", namespace
        )
        name = body.metadata.name
        await self._store_object(namespace, "NetworkPolicy", name, body)

    async def read_namespaced_network_policy(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Pod:
        """Read a network policy object.

        Parameters
        ----------
        name
            Name of the network policy.
        namespace
            Namespace of the network policy.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1NetworkPolicy
            Network policy object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the network policy was not found.
        """
        self._maybe_error("read_namespaced_network_policy", name, namespace)
        return self._get_object(namespace, "NetworkPolicy", name)

    # NODE API

    async def list_node(
        self,
        *,
        label_selector: str | None = None,
        _request_timeout: float | None = None,
    ) -> V1NodeList:
        """List node information.

        Parameters
        ----------
        label_selector
            Which objects to retrieve. All labels must match.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1NodeList
            The node information previouslyl stored with `set_nodes_for_test`,
            if any.
        """
        self._maybe_error("list_node")
        nodes = [
            n
            for n in self._nodes
            if _check_labels(n.metadata.labels, label_selector)
        ]
        return V1NodeList(items=nodes)

    # PERSISTENTVOLUME API

    async def create_persistent_volume(
        self,
        body: V1PersistentVolume,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a persistent volume.

        Parameters
        ----------
        body
            Persistent volume to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the persistent volume already
            exists.
        """
        self._maybe_error("create_persistent_volume", body)
        self._update_metadata(body, "v1", "PersistentVolume")
        await self._store_cluster_object(
            "PersistentVolume", body.metadata.name, body
        )

    async def delete_persistent_volume(
        self,
        name: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete a persistent volume object.

        Parameters
        ----------
        name
            Name of persistent volume to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the ingress was not found.
        """
        self._maybe_error("delete_persistent_volume", name)
        return self._delete_cluster_object(
            "PersistentVolume", name, propagation_policy
        )

    async def list_persistent_volume(
        self,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1PersistentVolumeList | Mock:
        """List persistent volume objects.

        This does support watches.

        Parameters
        ----------
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            persistent volume claim name and only persistent volume claims
            matching that name will be returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1PersistentVolumeList
            List of persistent volumes, when not called as a watch.
            If called as a watch, returns a mock ``aiohttp.Response``
            with a ``readline`` metehod that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        """
        self._maybe_error(
            "list_persistent_volume",
            field_selector,
        )
        if not watch:
            pvs = self._list_cluster_objects(
                "PersistentVolume",
                field_selector,
                label_selector,
            )
            return V1PersistentVolumeList(kind="PersistentVolume", items=pvs)

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._cluster_event_streams["PersistentVolume"]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def read_persistent_volume(
        self,
        name: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1PersistentVolume:
        """Read a persistent volume.

        Parameters
        ----------
        name
            Name of the persistent volume.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1PersistentVolume
            Persistent volume object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the persistent volume was not
            found.
        """
        self._maybe_error("read_persistent_volume", name)
        return self._get_cluster_object("PersistentVolume", name)

    # PERSISTENTVOLUMECLAIM API

    async def create_namespaced_persistent_volume_claim(
        self,
        namespace: str,
        body: V1PersistentVolumeClaim,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a persistent volume claim.

        Parameters
        ----------
        namespace
            Namespace in which to create the persistent volume claim.
        body
            Persistent volume claim to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the persistent volume claim already
            exists.
        """
        self._maybe_error(
            "create_namespaced_persistent_volume_claim", namespace, body
        )
        self._update_metadata(body, "v1", "PersistentVolumeClaim", namespace)
        await self._store_object(
            namespace, "PersistentVolumeClaim", body.metadata.name, body
        )

    async def delete_namespaced_persistent_volume_claim(
        self,
        name: str,
        namespace: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete a persistent volume claim object.

        Parameters
        ----------
        name
            Name of persistent volume claim to delete.
        namespace
            Namespace of persistent volume claim to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the ingress was not found.
        """
        self._maybe_error(
            "delete_namespaced_persistent_volume_claim", name, namespace
        )
        return self._delete_object(
            namespace, "PersistentVolumeClaim", name, propagation_policy
        )

    async def list_namespaced_persistent_volume_claim(
        self,
        namespace: str,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1IngressList | Mock:
        """List persistent volume claim objects in a namespace.

        This does support watches.

        Parameters
        ----------
        namespace
            Namespace of persistent volume claim to list.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            persistent volume claim name and only persistent volume claims
            matching that name will be returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1PersistentVolumeClaimList
            List of persistent volume claims in that namespace, when not
            called as a watch.  If called as a watch, returns a mock
            ``aiohttp.Response`` with a ``readline`` metehod that yields the
            events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error(
            "list_namespaced_persistent_volume_claim",
            namespace,
            field_selector,
        )
        if namespace not in self._objects:
            msg = f"Namespace {namespace} not found"
            raise ApiException(status=404, reason=msg)
        if not watch:
            pvcs = self._list_objects(
                namespace,
                "PersistentVolumeClaim",
                field_selector,
                label_selector,
            )
            return V1PersistentVolumeClaimList(
                kind="PersistentVolumeClaim", items=pvcs
            )

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace]["PersistentVolumeClaim"]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def read_namespaced_persistent_volume_claim(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1PersistentVolumeClaim:
        """Read a persistent volume claim.

        Parameters
        ----------
        name
            Name of the persistent volume claim.
        namespace
            Namespace of the persistent volume claim.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1PersistentVolumeClaim
            Persistent volume claim object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the persistent volume claim was not
            found.
        """
        self._maybe_error(
            "read_namespaced_persistent_volume_claim", name, namespace
        )
        return self._get_object(namespace, "PersistentVolumeClaim", name)

    # POD API

    async def create_namespaced_pod(
        self,
        namespace: str,
        body: V1Pod,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a pod object.

        If ``initial_pod_phase`` on the mock Kubernetes object is set to
        ``Running``, set the state to ``Running`` and generate a startup
        event. Otherwise, the status is set to whatever ``initial_pod_phase``
        is set to, and no event is generated.

        Parameters
        ----------
        namespace
            Namespace in which to create the pod.
        body
            Pod specification.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the pod already exists.
        """
        self._maybe_error("create_namespaced_pod", namespace, body)
        self._update_metadata(body, "v1", "Pod", namespace)
        body.status = V1PodStatus(phase=self.initial_pod_phase)
        body.status.start_time = current_datetime()
        await self._store_object(namespace, "Pod", body.metadata.name, body)

        # Post an event for the pod start if configured to mark pods as
        # started automatically.
        if self.initial_pod_phase == "Running":
            event = CoreV1Event(
                metadata=V1ObjectMeta(
                    name=f"{body.metadata.name}-start", namespace=namespace
                ),
                message=f"Pod {body.metadata.name} started",
                involved_object=V1ObjectReference(
                    kind="Pod", name=body.metadata.name, namespace=namespace
                ),
            )
            await self.create_namespaced_event(namespace, event)

    async def delete_namespaced_pod(
        self,
        name: str,
        namespace: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete a pod object.

        Parameters
        ----------
        name
            Name of pod to delete.
        namespace
            Namespace of pod to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the pod was not found.
        """
        self._maybe_error("delete_namespaced_pod", name, namespace)
        return self._delete_object(namespace, "Pod", name, propagation_policy)

    async def list_namespaced_pod(
        self,
        namespace: str,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1PodList | Mock:
        """List pod objects in a namespace.

        This does support watches.

        Parameters
        ----------
        namespace
            Namespace of pods to list.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            pod name and only pods matching that name will be returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1PodList or unittest.mock.Mock
            List of pods in that namespace, when not called as a watch. If
            called as a watch, returns a mock ``aiohttp.Response`` with a
            ``readline`` metehod that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error("list_namespaced_pod", namespace, field_selector)
        if namespace not in self._objects:
            msg = f"Namespace {namespace} not found"
            raise ApiException(status=404, reason=msg)
        if not watch:
            pods = self._list_objects(
                namespace, "Pod", field_selector, label_selector
            )
            return V1PodList(kind="Pod", items=pods)

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace]["Pod"]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def patch_namespaced_pod_status(
        self,
        name: str,
        namespace: str,
        body: list[dict[str, Any]],
        *,
        _request_timeout: float | None = None,
    ) -> V1Secret:
        """Patch the status of a pod object.

        Parameters
        ----------
        name
            Name of pod object.
        namespace
            Namespace of secret object.
        body
            Patches to apply. Only patches with ``op`` of ``replace`` are
            supported, and only with ``path`` of ``/status/phase``.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Pod
            Patched pod object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the secret does not exist.
        AssertionError
            Raised if any other type of patch was provided.
        """
        self._maybe_error("patch_namespaced_pod", name, namespace)
        pod = copy.deepcopy(self._get_object(namespace, "Pod", name))
        for change in body:
            assert change["op"] == "replace"
            assert change["path"] == "/status/phase"
            pod.status.phase = change["value"]
        await self._store_object(namespace, "Pod", name, pod, replace=True)
        return pod

    async def read_namespaced_pod(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Pod:
        """Read a pod object.

        Parameters
        ----------
        name
            Name of the pod.
        namespace
            Namespace of the pod.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Pod
            Pod object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the pod was not found.
        """
        self._maybe_error("read_namespaced_pod", name, namespace)
        return self._get_object(namespace, "Pod", name)

    async def read_namespaced_pod_status(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Pod:
        """Read the status of a pod.

        Parameters
        ----------
        name
            Name of the pod.
        namespace
            Namespace of the pod.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Pod
            Pod object. The Kubernetes API returns a ``V1Pod`` rather than, as
            expected, a ``V1PodStatus``. Presumably the acutal API populates
            only the status portion, but we return the whole pod for testing
            purposes since it shouldn't matter.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the pod was not found.
        """
        self._maybe_error("read_namespaced_pod_status", name, namespace)
        return self._get_object(namespace, "Pod", name)

    # RESOURCEQUOTA API

    async def create_namespaced_resource_quota(
        self,
        namespace: str,
        body: V1ResourceQuota,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a resource quota object.

        Parameters
        ----------
        namespace
            Namespace in which to create the object.
        body
            Object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_resource_quota", namespace, body)
        self._update_metadata(body, "v1", "ResourceQuota", namespace)
        name = body.metadata.name
        await self._store_object(namespace, "ResourceQuota", name, body)

    async def read_namespaced_resource_quota(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1ResourceQuota:
        """Read a resource quota object.

        Parameters
        ----------
        name
            Name of object.
        namespace
            Namespace of object.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1ResourceQuota
            Corresponding object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the resource quota was not found.
        """
        self._maybe_error("read_namespaced_resource_quota", name, namespace)
        return self._get_object(namespace, "ResourceQuota", name)

    # SECRETS API

    async def create_namespaced_secret(
        self,
        namespace: str,
        body: V1Secret,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a secret object.

        Parameters
        ----------
        namespace
            Namespace in which to create the object.
        body
            Object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_secret", namespace, body)
        self._update_metadata(body, "v1", "Secret", namespace)
        await self._store_object(namespace, "Secret", body.metadata.name, body)

    async def patch_namespaced_secret(
        self,
        name: str,
        namespace: str,
        body: list[dict[str, Any]],
        *,
        _request_timeout: float | None = None,
    ) -> V1Secret:
        """Patch a secret object.

        Parameters
        ----------
        name
            Name of secret object.
        namespace
            Namespace of secret object.
        body
            Patches to apply. Only patches with ``op`` of ``replace`` are
            supported, and only with ``path`` of either
            ``/metadata/annotations`` or ``/metadata/labels``.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Secret
            Patched secret object.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the secret does not exist.
        AssertionError
            Raised if any other type of patch was provided.
        """
        self._maybe_error("patch_namespaced_secret", name, namespace)
        secret = copy.deepcopy(self._get_object(namespace, "Secret", name))
        for change in body:
            assert change["op"] == "replace"
            if change["path"] == "/metadata/annotations":
                secret.metadata.annotations = change["value"]
            elif change["path"] == "/metadata/labels":
                secret.metadata.labels = change["value"]
            else:
                raise AssertionError(f"unsupported path {change['path']}")
        await self._store_object(
            namespace, "Secret", name, secret, replace=True
        )

    async def read_namespaced_secret(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Secret:
        """Read a secret.

        Parameters
        ----------
        name
            Name of secret.
        namespace
            Namespace of secret.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Secret
            Requested secret.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the secret does not exist.
        """
        self._maybe_error("read_namespaced_secret", name, namespace)
        return self._get_object(namespace, "Secret", name)

    async def replace_namespaced_secret(
        self,
        name: str,
        namespace: str,
        body: V1Secret,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Replace a secret.

        Parameters
        ----------
        name
            Name of secret.
        namespace
            Namespace of secret.
        body
            New body of secret.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the secret does not exist.
        """
        self._maybe_error("replace_namespaced_secret", namespace, body)
        await self._store_object(namespace, "Secret", name, body, replace=True)

    # SERVICE API

    async def create_namespaced_service(
        self,
        namespace: str,
        body: V1Service,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a service object.

        Parameters
        ----------
        namespace
            Namespace in which to create the object.
        body
            Object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_service", namespace, body)
        self._update_metadata(body, "v1", "Service", namespace)
        await self._store_object(
            namespace, "Service", body.metadata.name, body
        )

    async def delete_namespaced_service(
        self,
        name: str,
        namespace: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete a service object.

        Parameters
        ----------
        name
            Name of service to delete.
        namespace
            Namespace of service to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the service was not found.
        """
        self._maybe_error("delete_namespaced_service", name, namespace)
        return self._delete_object(
            namespace, "Service", name, propagation_policy
        )

    async def list_namespaced_service(
        self,
        namespace: str,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1ServiceList | Mock:
        """List service objects in a namespace.

        This does support watches.

        Parameters
        ----------
        namespace
            Namespace of services to list.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            service name and only services matching that name will be returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1ServiceList or unittest.mock.Mock
            List of services in that namespace, when not called as a watch. If
            called as a watch, returns a mock ``aiohttp.Response`` with a
            ``readline`` metehod that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error("list_namespaced_service", namespace, field_selector)
        if namespace not in self._objects:
            msg = f"Namespace {namespace} not found"
            raise ApiException(status=404, reason=msg)
        if not watch:
            services = self._list_objects(
                namespace, "Service", field_selector, label_selector
            )
            return V1ServiceList(kind="Service", items=services)

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace]["Service"]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def read_namespaced_service(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Service:
        """Read a service.

        Parameters
        ----------
        name
            Name of service.
        namespace
            Namespace of service.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Service
            Requested service.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the service does not exist.
        """
        self._maybe_error("read_namespaced_service", name, namespace)
        return self._get_object(namespace, "Service", name)

    # SERVICEACCOUNT API

    async def create_namespaced_service_account(
        self,
        namespace: str,
        body: V1ServiceAccount,
        *,
        _request_timeout: float | None = None,
    ) -> None:
        """Create a service account object.

        Parameters
        ----------
        namespace
            Namespace in which to create the object.
        body
            Object to create.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 409 status if the object already exists.
        """
        self._maybe_error("create_namespaced_service_account", namespace, body)
        self._update_metadata(body, "v1", "ServiceAccount", namespace)
        await self._store_object(
            namespace, "ServiceAccount", body.metadata.name, body
        )

    async def delete_namespaced_service_account(
        self,
        name: str,
        namespace: str,
        *,
        grace_period_seconds: int | None = None,
        propagation_policy: str = "Foreground",
        body: V1DeleteOptions | None = None,
        _request_timeout: float | None = None,
    ) -> V1Status:
        """Delete a service account object.

        Parameters
        ----------
        name
            Name of service account to delete.
        namespace
            Namespace of service account to delete.
        grace_period_seconds
            Grace period for object deletion (currently ignored).
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        body
            Delete options (currently ignored).
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            Success status.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the service was not found.
        """
        self._maybe_error("delete_namespaced_service_account", name, namespace)
        return self._delete_object(
            namespace, "ServiceAccount", name, propagation_policy
        )

    async def list_namespaced_service_account(
        self,
        namespace: str,
        *,
        field_selector: str | None = None,
        label_selector: str | None = None,
        resource_version: str | None = None,
        timeout_seconds: int | None = None,
        watch: bool = False,
        _preload_content: bool = True,
        _request_timeout: float | None = None,
    ) -> V1ServiceList | Mock:
        """List service account objects in a namespace.

        This does support watches.

        Parameters
        ----------
        namespace
            Namespace of service accounts to list.
        field_selector
            Only ``metadata.name=...`` is supported. It is parsed to find the
            service name and only services matching that name will be returned.
        label_selector
            Which objects to retrieve. All labels must match.
        resource_version
            Where to start in the event stream when performing a watch. If
            `None`, starts with the next change.
        timeout_seconds
            How long to return events for before exiting when performing a
            watch.
        watch
            Whether to act as a watch.
        _preload_content
            Verified to be `False` when performing a watch.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1ServiceAccountList or unittest.mock.Mock
            List of service accounts in that namespace, when not called as a
            watch. If called as a watch, returns a mock ``aiohttp.Response``
            with a ``readline`` metehod that yields the events.

        Raises
        ------
        AssertionError
            Some other ``field_selector`` was provided.
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the namespace does not exist.
        """
        self._maybe_error(
            "list_namespaced_service_account", namespace, field_selector
        )
        if namespace not in self._objects:
            msg = f"Namespace {namespace} not found"
            raise ApiException(status=404, reason=msg)
        if not watch:
            accounts = self._list_objects(
                namespace, "ServiceAccount", field_selector, label_selector
            )
            return V1ServiceAccountList(kind="ServiceAccount", items=accounts)

        # All watches must not preload content since we're returning raw JSON.
        # This is done by the Kubernetes API Watch object.
        assert not _preload_content

        # Return the mock response expected by the Kubernetes API.
        stream = self._event_streams[namespace]["ServiceAccount"]
        return stream.build_watch_response(
            resource_version,
            timeout_seconds,
            field_selector=field_selector,
            label_selector=label_selector,
        )

    async def read_namespaced_service_account(
        self,
        name: str,
        namespace: str,
        *,
        _request_timeout: float | None = None,
    ) -> V1Service:
        """Read a service account.

        Parameters
        ----------
        name
            Name of service account.
        namespace
            Namespace of service account.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1ServiceAccount
            Requested service account.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the service does not exist.
        """
        self._maybe_error("read_namespaced_service_account", name, namespace)
        return self._get_object(namespace, "ServiceAccount", name)

    # Internal helper functions.

    def _delete_object(
        self, namespace: str, key: str, name: str, propagation_policy: str
    ) -> V1Status:
        """Delete an object from internal data structures.

        Parameters
        ----------
        namespace
            Namespace from which to delete an object.
        key
            Key under which the object is stored (usually the kind).
        name
            Name of the object.
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            200 return status if the object was found and deleted.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with a 404 status if the object is not found.
        """
        if propagation_policy not in ("Foreground", "Background", "Orphan"):
            msg = f"Invalid propagation_policy {propagation_policy}"
            raise AssertionError(msg)
        obj = self._get_object(namespace, key, name)
        del self._objects[namespace][key][name]
        stream = self._event_streams[namespace][key]
        if isinstance(obj, dict):
            stream.add_custom_event("DELETED", obj)
        else:
            stream.add_event("DELETED", obj)
        return V1Status(code=200)

    def _delete_cluster_object(
        self, key: str, name: str, propagation_policy: str
    ) -> V1Status:
        """Delete a cluster-scoped object from internal data structures.

        Parameters
        ----------
        key
            Key under which the object is stored (usually the kind).
        name
            Name of the object.
        propagation_policy
            Propagation policy for deletion. Has no effect on the mock.
        _request_timeout
            Ignored, accepted for compatibility with the Kubernetes API.

        Returns
        -------
        kubernetes_asyncio.client.V1Status
            200 return status if the object was found and deleted.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with a 404 status if the object is not found.
        """
        if propagation_policy not in ("Foreground", "Background", "Orphan"):
            msg = f"Invalid propagation_policy {propagation_policy}"
            raise AssertionError(msg)
        obj = self._get_cluster_object(key, name)
        del self._cluster_objects[key][name]
        stream = self._cluster_event_streams[key]
        if isinstance(obj, dict):
            stream.add_custom_event("DELETED", obj)
        else:
            stream.add_event("DELETED", obj)
        return V1Status(code=200)

    def _get_object(self, namespace: str, key: str, name: str) -> Any:
        """Retrieve an object from internal data structures.

        Parameters
        ----------
        namespace
            Namespace from which to delete an object.
        key
            Key under which the object is stored (usually the kind).
        name
            Name of the object.

        Returns
        -------
        Any
            Object if found.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with a 404 status if the object is not found.
        """
        if namespace not in self._objects:
            reason = f"{namespace}/{name} not found"
            raise ApiException(status=404, reason=reason)
        if name not in self._objects[namespace].get(key, {}):
            reason = f"{namespace}/{name} not found"
            raise ApiException(status=404, reason=reason)
        return self._objects[namespace][key][name]

    def _get_cluster_object(self, key: str, name: str) -> Any:
        """Retrieve a cluster-scoped object from internal data structures.

        Parameters
        ----------
        key
            Key under which the object is stored (usually the kind).
        name
            Name of the object.

        Returns
        -------
        Any
            Object if found.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with a 404 status if the object is not found.
        """
        if name not in self._cluster_objects.get(key, {}):
            reason = f"{name} not found"
            raise ApiException(status=404, reason=reason)
        return self._cluster_objects[key][name]

    def _list_objects(
        self,
        namespace: str,
        key: str,
        field_selector: str | None,
        label_selector: str | None,
    ) -> list[Any]:
        """List objects, possibly with selector restrictions.

        Parameters
        ----------
        namespace
            Namespace in which to list objects.
        key
            Key under which the object is stored (usually the kind).
        field_selector
            If present, only ``metadata.name=...`` is supported. It is parsed
            to find the object name and only an object matching that name will
            be returned.
        label_selector
            Which matching objects to retrieve by label. All labels must
            match.

        Returns
        -------
        list
            List of matching objects.
        """
        if key not in self._objects[namespace]:
            return []

        # If there is a field selector, only name selectors are supported and
        # we should retrieve the object by name.
        if field_selector:
            match = re.match(r"metadata\.name=(.*)$", field_selector)
            if not match or not match.group(1):
                msg = f"Field selector {field_selector} not supported"
                raise ValueError(msg)
            try:
                obj = self._get_object(namespace, key, match.group(1))
                if _check_labels(obj.metadata.labels, label_selector):
                    return [obj]
                else:
                    return []
            except ApiException:
                return []

        # Otherwise, construct the list of all objects matching the label
        # selector.
        return [
            o
            for o in self._objects[namespace][key].values()
            if _check_labels(o.metadata.labels, label_selector)
        ]

    def _list_cluster_objects(
        self,
        key: str,
        field_selector: str | None,
        label_selector: str | None,
    ) -> list[Any]:
        """List cluster-scoped objects, possibly with selector restrictions.

        Parameters
        ----------
        key
            Key under which the object is stored (usually the kind).
        field_selector
            If present, only ``metadata.name=...`` is supported. It is parsed
            to find the object name and only an object matching that name will
            be returned.
        label_selector
            Which matching objects to retrieve by label. All labels must
            match.

        Returns
        -------
        list
            List of matching objects.
        """
        if key not in self._cluster_objects:
            return []

        # If there is a field selector, only name selectors are supported and
        # we should retrieve the object by name.
        if field_selector:
            match = re.match(r"metadata\.name=(.*)$", field_selector)
            if not match or not match.group(1):
                msg = f"Field selector {field_selector} not supported"
                raise ValueError(msg)
            try:
                obj = self._get_cluster_object(key, match.group(1))
                if _check_labels(obj.metadata.labels, label_selector):
                    return [obj]
                else:
                    return []
            except ApiException:
                return []

        # Otherwise, construct the list of all objects matching the label
        # selector.
        return [
            o
            for o in self._cluster_objects[key].values()
            if _check_labels(o.metadata.labels, label_selector)
        ]

    def _maybe_error(self, method: str, *args: Any) -> None:
        """Call the error callback if one is registered.

        This is a separate helper function to avoid using class method call
        syntax.
        """
        if self.error_callback:
            callback = self.error_callback
            callback(method, *args)

    async def _store_object(
        self,
        namespace: str,
        key: str,
        name: str,
        obj: Any,
        *,
        replace: bool = False,
    ) -> None:
        """Store an object in internal data structures.

        Parameters
        ----------
        namespace
            Namespace in which to store the object.
        key
            Kind of the object for most objects, or a more complex key for
            custom objects.
        name
            Name of object.
        obj
            Object to store.
        replace
            If `True`, the object must already exist, to mirror the Kubernetes
            replace semantices. If `False`, a conflict error is raised if the
            object already exists.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the object does not exist and ``replace``
            was `True`, and with 409 status if the object does exist and
            ``replace`` was `False`.
        """
        if replace:
            self._get_object(namespace, key, name)
        else:
            if namespace not in self._objects:
                self._objects[namespace] = {}
            if key not in self._objects[namespace]:
                self._objects[namespace][key] = {}
            if name in self._objects[namespace][key]:
                msg = f"{namespace}/{name} exists"
                raise ApiException(status=409, reason=msg)
        if key == "Namespace":
            stream = self._namespace_stream
        else:
            stream = self._event_streams[namespace][key]
        if isinstance(obj, dict):
            obj["metadata"]["resourceVersion"] = stream.next_resource_version
            self._objects[namespace][key][name] = obj
            stream.add_custom_event("MODIFIED" if replace else "ADDED", obj)
        else:
            obj.metadata.resource_version = stream.next_resource_version
            self._objects[namespace][key][name] = obj
            stream.add_event("MODIFIED" if replace else "ADDED", obj)

        # Handling creation hooks by kind for custom objects is a bit
        # complicated because we only have the key, which has the API path for
        # the custom object but not the capitalization of the kind. However,
        # it should be registered in self._custom_kinds by this point, so we
        # can find it there.
        kind = key
        for candidate, full_key in self._custom_kinds.items():
            if full_key == key:
                kind = candidate
        if kind in self._create_hooks:
            hook = self._create_hooks[kind]
            await hook(obj)

    async def _store_cluster_object(
        self,
        key: str,
        name: str,
        obj: Any,
        *,
        replace: bool = False,
    ) -> None:
        """Store a cluster-scoped object in internal data structures.

        Parameters
        ----------
        key
            Kind of the object for most objects, or a more complex key for
            custom objects.
        name
            Name of object.
        obj
            Object to store.
        replace
            If `True`, the object must already exist, to mirror the Kubernetes
            replace semantices. If `False`, a conflict error is raised if the
            object already exists.

        Raises
        ------
        kubernetes_asyncio.client.ApiException
            Raised with 404 status if the object does not exist and ``replace``
            was `True`, and with 409 status if the object does exist and
            ``replace`` was `False`.
        """
        if replace:
            self._get_cluster_object(key, name)
        else:
            if key not in self._cluster_objects:
                self._cluster_objects[key] = {}
            if name in self._cluster_objects[key]:
                msg = f"{name} exists"
                raise ApiException(status=409, reason=msg)
        stream = self._cluster_event_streams[key]
        if isinstance(obj, dict):
            obj["metadata"]["resourceVersion"] = stream.next_resource_version
            self._cluster_objects[key][name] = obj
            stream.add_custom_event("MODIFIED" if replace else "ADDED", obj)
        else:
            obj.metadata.resource_version = stream.next_resource_version
            self._cluster_objects[key][name] = obj
            stream.add_event("MODIFIED" if replace else "ADDED", obj)

    def _update_metadata(
        self,
        body: Any,
        api_version: str,
        kind: str,
        namespace: str | None = None,
    ) -> None:
        """Check and potentially update the metadata of a stored object.

        The Kubernetes API allows the ``api_version``, ``kind``, and
        ``metadata.namespace`` attributes to be omitted since they can be
        determined from the method called or its parameters. Implement the
        same logic, but if they are provided in the object, require that they
        be correct (match the parameters inferred from the call).

        Parameters
        ----------
        body
            Object being stored. Updated in place with namespace, kind, and
            API version information.
        api_version
            Expected API version of object.
        kind
            Expected kind of object.
        namespace
            Namespace in which it is being stored, or `None` for objects that
            are not namespaced.

        Raises
        ------
        AssertionError
            Raised if the namespace, kind, or API version was provided but
            didn't match the expected value.
        """
        if body.api_version:
            assert body.api_version == api_version
        else:
            body.api_version = api_version
        if body.kind:
            assert body.kind == kind
        else:
            body.kind = kind
        if body.metadata.namespace:
            assert body.metadata.namespace == namespace
        else:
            body.metadata.namespace = namespace


def patch_kubernetes() -> Iterator[MockKubernetesApi]:
    """Replace the Kubernetes API with a mock class.

    Returns
    -------
    MockKubernetesApi
        The mock Kubernetes API object.

    Notes
    -----
    This function will mock out the Kuberentes library configuration API in a
    manner compatible with `safir.kubernetes.initialize_kubernetes`, ensuring
    that it does nothing during test.  It will replace the Kubernetes
    ``ApiClient`` object with a `~unittest.mock.MagicMock` and then redirect
    ``CoreV1Api`` and ``CustomObjectsApi`` to a `MockKubernetesApi` instance.

    To use this mock successfully, you must not import ``ApiClient``,
    ``CoreV1Api``, or ``CustomObjectsApi`` directly into the local namespace,
    or they will not be correctly patched.  Instead, use:

    .. code-block:: python

       from kubernetes_asyncio import client

    and then use ``client.ApiClient`` and so forth.

    Examples
    --------
    Normally this should be called from a fixture in ``tests/conftest.py``
    such as the following:

    .. code-block:: python

       from safir.testing.kubernetes import MockKubernetesApi, patch_kubernetes


       @pytest.fixture
       def mock_kubernetes() -> Iterator[MockKubernetesApi]:
           yield from patch_kubernetes()
    """
    mock_api = MockKubernetesApi()
    with patch.object(config, "load_incluster_config"):
        patchers = []
        for api in (
            "BatchV1Api",
            "CoreV1Api",
            "CustomObjectsApi",
            "NetworkingV1Api",
        ):
            patcher = patch.object(client, api)
            mock_class = patcher.start()
            mock_class.return_value = mock_api
            patchers.append(patcher)
        with patch.object(client.ApiClient, "request"):
            os.environ["KUBERNETES_PORT"] = "tcp://10.0.0.1:443"
            yield mock_api
            del os.environ["KUBERNETES_PORT"]
        for patcher in patchers:
            patcher.stop()



"""Tests for Kubernetes utilities."""

from __future__ import annotations

import pytest
from kubernetes_asyncio import config

from safir.kubernetes import initialize_kubernetes
from safir.testing.kubernetes import MockKubernetesApi


@pytest.mark.asyncio
async def test_initialize(mock_kubernetes: MockKubernetesApi) -> None:
    await initialize_kubernetes()
    assert config.load_incluster_config.call_count == 1



"""Test fixtures."""

from __future__ import annotations

from collections.abc import (
    AsyncGenerator,
    Callable,
    Generator,
    Iterator,
    Sequence,
)
from datetime import timedelta, timezone
from typing import Any

import pytest
import pytest_asyncio
import respx
import structlog
from aiokafka.admin.client import NewTopic
from arq.connections import ArqRedis, RedisSettings
from arq.constants import expires_extra_ms
from arq.cron import CronJob
from arq.jobs import Deserializer, Serializer
from arq.typing import SecondsTimedelta, StartupShutdown, WorkerCoroutine
from arq.worker import Function, Worker
from redis.asyncio import Redis
from structlog.testing import LogCapture
from testcontainers.postgres import PostgresContainer
from testcontainers.redis import RedisContainer

from safir.kafka import FastStreamErrorHandler
from safir.metrics import (
    EventManager,
    EventsConfiguration,
    KafkaMetricsConfiguration,
)
from safir.sentry import (
    before_send_handler,
    fingerprint_env_handler,
    sentry_exception_handler,
)
from safir.testing.gcs import MockStorageClient, patch_google_storage
from safir.testing.kubernetes import MockKubernetesApi, patch_kubernetes
from safir.testing.sentry import (
    Captured,
    capture_events_fixture,
    sentry_init_fixture,
)
from safir.testing.slack import MockSlackWebhook, mock_slack_webhook

from .support.kafka import (
    KafkaClients,
    KafkaStack,
    make_kafka_clients,
    make_kafka_stack,
)
from .support.metrics import MetricsStack, make_metrics_stack


@pytest.fixture
def log_output() -> LogCapture:
    """Capture Structlog logs to assert against."""
    return LogCapture()


@pytest.fixture(autouse=True)
def configure_structlog(log_output: LogCapture) -> None:
    structlog.configure(processors=[log_output])


@pytest.fixture(scope="session")
def session_kafka_stack(
    tmp_path_factory: pytest.TempPathFactory,
) -> Generator[KafkaStack]:
    kafka_cert_path = tmp_path_factory.mktemp("kafka-certs")
    with make_kafka_stack(kafka_cert_path=kafka_cert_path) as stack:
        yield stack


@pytest.fixture
def kafka_stack(
    session_kafka_stack: KafkaStack,
) -> KafkaStack:
    session_kafka_stack.kafka_container.reset()
    session_kafka_stack.schema_registry_container.reset()
    return session_kafka_stack


@pytest_asyncio.fixture
async def kafka_clients(
    kafka_stack: KafkaStack,
) -> AsyncGenerator[KafkaClients]:
    async with make_kafka_clients(kafka_stack) as clients:
        yield clients


@pytest_asyncio.fixture
async def event_manager(
    kafka_stack: KafkaStack,
) -> AsyncGenerator[EventManager]:
    """Provide an event manager and create a matching Kafka topic."""
    config = KafkaMetricsConfiguration(
        application="testapp",
        enabled=True,
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=kafka_stack.kafka_connection_settings,
        schema_manager=kafka_stack.schema_manager_settings,
    )

    manager = config.make_manager()
    await manager.initialize()
    topic = NewTopic(
        name="what.ever.testapp",
        num_partitions=1,
        replication_factor=1,
    )
    await manager._admin_client.create_topics([topic])
    yield manager
    await manager.aclose()


@pytest.fixture
def test_scoped_kafka_stack(
    tmp_path_factory: pytest.TempPathFactory,
) -> Generator[KafkaStack]:
    kafka_cert_path = tmp_path_factory.mktemp("kafka-certs")
    with make_kafka_stack(
        kafka_cert_path=kafka_cert_path, constant_host_ports=True
    ) as stack:
        yield stack


@pytest_asyncio.fixture
async def test_scoped_kafka_clients(
    test_scoped_kafka_stack: KafkaStack,
) -> AsyncGenerator[KafkaClients]:
    async with make_kafka_clients(test_scoped_kafka_stack) as clients:
        yield clients


@pytest_asyncio.fixture
async def resiliency_metrics_stack(
    test_scoped_kafka_stack: KafkaStack,
    test_scoped_kafka_clients: KafkaClients,
) -> AsyncGenerator[MetricsStack]:
    async with make_metrics_stack(
        kafka_stack=test_scoped_kafka_stack,
        kafka_clients=test_scoped_kafka_clients,
        kafka_max_batch_size=1,
    ) as stack:
        yield stack


@pytest.fixture
def faststream_error_handler() -> FastStreamErrorHandler:
    """Create a FastStreamErrorHandler instance.

    This can be used in a test function to configure Slack error reporting for
    a FastStream broker or router.
    """
    return FastStreamErrorHandler()


@pytest.fixture(scope="session")
def database_password() -> str:
    return "INSECURE@%PASSWORD/"


@pytest.fixture(scope="session")
def database_url(database_password: str) -> Iterator[str]:
    """Start a PostgreSQL database and return a URL for it."""
    with PostgresContainer(
        driver="asyncpg", username="safir", password=database_password
    ) as postgres:
        yield postgres.get_connection_url()


@pytest.fixture
def mock_gcs() -> Iterator[MockStorageClient]:
    yield from patch_google_storage(
        expected_expiration=timedelta(hours=1), bucket_name="some-bucket"
    )


@pytest.fixture
def mock_kubernetes() -> Iterator[MockKubernetesApi]:
    yield from patch_kubernetes()


@pytest.fixture
def mock_slack(respx_mock: respx.Router) -> MockSlackWebhook:
    return mock_slack_webhook("https://example.com/slack", respx_mock)


@pytest.fixture(scope="session")
def redis() -> Iterator[RedisContainer]:
    """Start a Redis container."""
    with RedisContainer() as redis:
        yield redis


@pytest_asyncio.fixture
async def redis_client(redis: RedisContainer) -> AsyncGenerator[Redis]:
    """Create a Redis client for testing.

    This must be done separately for each test since it's tied to the per-test
    event loop, and therefore must be separated from the session-shared Redis
    server container.
    """
    host = redis.get_container_host_ip()
    port = redis.get_exposed_port(6379)
    client = Redis(host=host, port=port, db=0)
    yield client
    await client.aclose()


@pytest_asyncio.fixture
async def arq_redis(redis: RedisContainer) -> AsyncGenerator[ArqRedis]:
    """Create ArqRedis connected to the global redis container."""
    arq_redis_ = ArqRedis(
        host=redis.get_container_host_ip(),
        port=redis.get_exposed_port(6379),
    )
    await arq_redis_.flushall()
    yield arq_redis_
    await arq_redis_.aclose(close_connection_pool=True)


@pytest_asyncio.fixture
async def create_arq_worker(arq_redis: ArqRedis) -> AsyncGenerator[Callable]:
    """Make a function to make an arq worker connected to the global redis.

    This is the fixture in the arq tests:
    https://github.com/python-arq/arq/blob/main/tests/conftest.py
    """
    worker_: Worker | None = None

    def create(
        functions: Sequence[Function | WorkerCoroutine],
        *,
        queue_name: str | None = None,
        cron_jobs: Sequence[CronJob] | None = None,
        redis_settings: RedisSettings | None = None,
        redis_pool: ArqRedis | None = arq_redis,
        burst: bool = True,
        on_startup: StartupShutdown | None = None,
        on_shutdown: StartupShutdown | None = None,
        on_job_start: StartupShutdown | None = None,
        on_job_end: StartupShutdown | None = None,
        after_job_end: StartupShutdown | None = None,
        handle_signals: bool = True,
        job_completion_wait: int = 0,
        max_jobs: int = 10,
        job_timeout: SecondsTimedelta = 300,
        keep_result: SecondsTimedelta = 3600,
        keep_result_forever: bool = False,
        poll_delay: int = 0,
        queue_read_limit: int | None = None,
        max_tries: int = 5,
        health_check_interval: SecondsTimedelta = 3600,
        health_check_key: str | None = None,
        ctx: dict[Any, Any] | None = None,
        retry_jobs: bool = True,
        allow_abort_jobs: bool = False,
        max_burst_jobs: int = -1,
        job_serializer: Serializer | None = None,
        job_deserializer: Deserializer | None = None,
        expires_extra_ms: int = expires_extra_ms,
        timezone: timezone | None = None,
        log_results: bool = True,
    ) -> Worker:
        nonlocal worker_
        worker_ = Worker(
            functions=functions,
            queue_name=queue_name,
            cron_jobs=cron_jobs,
            redis_settings=redis_settings,
            redis_pool=redis_pool,
            burst=burst,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            on_job_start=on_job_start,
            on_job_end=on_job_end,
            after_job_end=after_job_end,
            handle_signals=handle_signals,
            job_completion_wait=job_completion_wait,
            max_jobs=max_jobs,
            job_timeout=job_timeout,
            keep_result=keep_result,
            keep_result_forever=keep_result_forever,
            poll_delay=poll_delay,
            queue_read_limit=queue_read_limit,
            max_tries=max_tries,
            health_check_interval=health_check_interval,
            health_check_key=health_check_key,
            ctx=ctx,
            retry_jobs=retry_jobs,
            allow_abort_jobs=allow_abort_jobs,
            max_burst_jobs=max_burst_jobs,
            job_serializer=job_serializer,
            job_deserializer=job_deserializer,
            expires_extra_ms=expires_extra_ms,
            timezone=timezone,
            log_results=log_results,
        )
        return worker_

    yield create

    if worker_:
        await worker_.close()


@pytest.fixture
def sentry_fingerprint_items(
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[Captured]:
    """Mock sentry transport and add env to event fingerprints."""
    with sentry_init_fixture() as init:
        init(
            environment="some_env",
            traces_sample_rate=1.0,
            before_send=fingerprint_env_handler,
        )
        events = capture_events_fixture(monkeypatch)
        yield events()


@pytest.fixture
def sentry_exception_items(
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[Captured]:
    """Mock sentry transport and add custom exception info."""
    with sentry_init_fixture() as init:
        init(
            environment="some_env",
            traces_sample_rate=1.0,
            before_send=sentry_exception_handler,
        )
        events = capture_events_fixture(monkeypatch)
        yield events()


@pytest.fixture
def sentry_combo_items(
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[Captured]:
    """Mock sentry transport and add all recommended before_send processing."""
    with sentry_init_fixture() as init:
        init(
            environment="some_env",
            traces_sample_rate=1.0,
            before_send=before_send_handler,
        )
        events = capture_events_fixture(monkeypatch)
        yield events()



"""Tests for UWS exception classes."""

from __future__ import annotations

import pickle

from safir.arq.uws import (
    WorkerError,
    WorkerFatalError,
    WorkerTransientError,
    WorkerUsageError,
)
from safir.uws._exceptions import TaskError


def test_pickle() -> None:
    """Test that the `TaskError` exceptions can be pickled.

    Some care has to be taken in defining exceptions to ensure that they can
    be pickled and unpickled, since `BaseException` provides a ``__reduce__``
    implementation with somewhat unexpected properties. Make sure that this
    support doesn't regress, since arq uses pickle to convey errors from
    backend workers to anything that recovers their results.
    """

    def nonnegative(arg: int) -> int:
        if arg < 0:
            raise ValueError("Negative integers not supported")
        return arg

    def raise_exception(arg: int, exc_class: type[WorkerError]) -> None:
        try:
            nonnegative(arg)
        except Exception as e:
            raise exc_class(
                "some message", "some detail", add_traceback=True
            ) from e

    for exc_class in (
        WorkerFatalError,
        WorkerTransientError,
        WorkerUsageError,
    ):
        exc: WorkerError = exc_class("some message", "detail")
        pickled_exc = pickle.loads(pickle.dumps(exc))
        task_error = TaskError.from_worker_error(exc)
        pickled_task_error = TaskError.from_worker_error(pickled_exc)
        job_error = task_error.to_job_error()
        assert job_error == pickled_task_error.to_job_error()
        if exc_class == WorkerUsageError:
            assert task_error.slack_ignore
        else:
            assert not task_error.slack_ignore

        # Try with tracebacks.
        try:
            raise_exception(-1, exc_class)
        except WorkerError as e:
            exc = e
        assert exc.traceback
        assert "nonnegative" in exc.traceback
        assert exc_class.__name__ not in exc.traceback
        job_error = TaskError.from_worker_error(exc).to_job_error()
        assert job_error.detail
        assert "some detail\n\n" in job_error.detail
        assert "nonnegative" in job_error.detail
        pickled_exc = pickle.loads(pickle.dumps(exc))
        pickled_task_error = TaskError.from_worker_error(pickled_exc)
        assert job_error == pickled_task_error.to_job_error()



"""Configuration for arq_ clients and workers."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass

from arq.connections import RedisSettings
from arq.constants import default_queue_name as arq_default_queue_name
from arq.cron import CronJob
from arq.typing import SecondsTimedelta, StartupShutdown, WorkerCoroutine
from arq.worker import Function
from pydantic import SecretStr
from pydantic_core import Url

__all__ = [
    "WorkerSettings",
    "build_arq_redis_settings",
]


def build_arq_redis_settings(
    url: Url, password: SecretStr | None
) -> RedisSettings:
    """Construct Redis settings for arq.

    Parameters
    ----------
    url
        Redis DSN.
    password
        Password for the Redis connection.

    Returns
    -------
    arq.connections.RedisSettings
        Settings for the arq Redis pool.

    Examples
    --------
    This function is normally used from a property in the application
    configuration. The application should usually use
    `~safir.pydantic.EnvRedisDsn` as the type for the Redis DSN.

    .. code-block:: python

       from arq.connections import RedisSettings
       from pydantic_settings import BaseSettings
       from safir.pydantic import EnvRedisDsn


       class Config(BaseSettings):
           arq_queue_url: EnvRedisDsn
           arq_queue_password: SecretStr | None

           @property
           def arq_redis_settings(self) -> RedisSettings:
               return build_arq_redis_settings(
                   self.arq_queue_url, self_arq_queue_password
               )
    """
    return RedisSettings(
        host=url.unicode_host() or "localhost",
        port=url.port or 6379,
        database=int(url.path.lstrip("/")) if url.path else 0,
        password=password.get_secret_value() if password else None,
    )


@dataclass
class WorkerSettings:
    """Configuration class for an arq worker.

    The arq command-line tool reads a class of the name ``WorkerSettings`` in
    the module it was given on the command line and turns its attributes into
    parameters to `arq.worker.Worker`. This dataclass represents the subset of
    the available settings that Safir applications have needed to date, as an
    aid for constructing that configuration object.
    """

    functions: Sequence[Function | WorkerCoroutine]
    """Coroutines to register as arq worker entry points."""

    redis_settings: RedisSettings
    """Redis configuration for arq."""

    job_completion_wait: SecondsTimedelta
    """How long to wait for jobs to complete before cancelling them."""

    job_timeout: SecondsTimedelta
    """Maximum timeout for all jobs."""

    max_jobs: int
    """Maximum number of jobs that can be run at one time."""

    allow_abort_jobs: bool = False
    """Whether to allow jobs to be aborted."""

    queue_name: str = arq_default_queue_name
    """Name of arq queue to listen to for jobs."""

    on_startup: StartupShutdown | None = None
    """Coroutine to run on startup."""

    on_shutdown: StartupShutdown | None = None
    """Coroutine to run on shutdown."""

    cron_jobs: Sequence[CronJob] | None = None
    """Cron jobs to run."""



"""Mocks and functions for testing services using the Safir UWS support."""

from ._assert import assert_job_summary_equal
from ._mocks import MockUWSJobRunner, MockWobbly, patch_wobbly

__all__ = [
    "MockUWSJobRunner",
    "MockWobbly",
    "assert_job_summary_equal",
    "patch_wobbly",
]



"""Logger dependency for FastAPI.

Provides a `structlog` logger as a FastAPI dependency.  The logger will
incorporate information from the request in its bound context.
"""

import uuid

import structlog
from fastapi import Request
from structlog.stdlib import BoundLogger

_logger_name: str | None = None
"""Name of the configured global logger.

When `configure_logging` is called, the name of the configured logger is
stored in this variable. It is used by the logger dependency to retrieve the
application's configured logger.

Only one configured logger is supported. Additional calls to
`configure_logging` change the stored logger name.

This variable is internal to Safir and should not be used outside of the
library.
"""

__all__ = ["LoggerDependency", "logger_dependency"]


class LoggerDependency:
    """Provides a structlog logger configured with request information.

    The following additional information will be included:

    * A UUID for the request
    * The method and URL of the request
    * The IP address of the client
    * The ``User-Agent`` header of the request, if any.

    The last three pieces of information will be added using naming consistent
    with the expectations of Google Log Explorer so that the request
    information will be liftedn into the appropriate JSON fields for complex
    log queries.
    """

    def __init__(self) -> None:
        self.logger: BoundLogger | None = None

    async def __call__(self, request: Request) -> BoundLogger:
        """Return a logger bound with request information.

        Returns
        -------
        structlog.stdlib.BoundLogger
            The bound logger.
        """
        if not self.logger:
            self.logger = structlog.get_logger(_logger_name)
            if not self.logger:
                msg = f"Unable to get logger for {_logger_name}"
                raise RuntimeError(msg)

        # Construct the httpRequest logging data (compatible with the format
        # expected by Google Log Explorer).
        request_data = {
            "requestMethod": request.method,
            "requestUrl": str(request.url),
        }
        if request.client:
            request_data["remoteIp"] = request.client.host
        user_agent = request.headers.get("User-Agent")
        if user_agent:
            request_data["userAgent"] = user_agent

        return self.logger.new(
            httpRequest=request_data,
            request_id=str(uuid.uuid4()),
        )


logger_dependency = LoggerDependency()
"""The dependency that will return the logger for the current request."""



"""Support classes for testing Alembic."""

from __future__ import annotations

from pydantic import SecretStr
from pydantic_core import Url
from pydantic_settings import BaseSettings, SettingsConfigDict
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from safir.logging import LogLevel
from safir.pydantic import EnvAsyncPostgresDsn

__all__ = [
    "BaseV1",
    "BaseV2",
    "UserV1",
    "UserV2",
    "config",
]


class Config(BaseSettings):
    """Just enough configuration to support the Alembic environments.

    When running Alembic via subprocess, set ``TEST_DATABASE_URL`` and
    ``TEST_DATABASE_PASSWORD`` to match the database that it should use.
    """

    model_config = SettingsConfigDict(env_prefix="TEST_", case_sensitive=False)

    database_url: EnvAsyncPostgresDsn = Url("postgresql://localhost/safir")
    database_password: SecretStr = SecretStr("INSECURE")
    log_level: LogLevel = LogLevel.DEBUG


config = Config()
"""Global config object used by Alembic."""


class BaseV1(DeclarativeBase):
    """First version of a database schema."""


class UserV1(BaseV1):
    """First version of a users table."""

    __tablename__ = "users"

    username: Mapped[str] = mapped_column(primary_key=True)
    name: Mapped[str | None]


class BaseV2(DeclarativeBase):
    """Second version of a database schema."""


class UserV2(BaseV2):
    """First version of a users table."""

    __tablename__ = "users"

    username: Mapped[str] = mapped_column(primary_key=True)
    name: Mapped[str | None]
    email: Mapped[str | None]



"""Configuration for the Kafka Schema Registry."""

from __future__ import annotations

from typing import TypedDict

from pydantic import AliasChoices, AnyUrl, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

try:
    from schema_registry.client import AsyncSchemaRegistryClient
except ImportError as e:
    raise ImportError(
        "The safir.kafka module requires the kafka extra. "
        "Install it with `pip install safir[kafka]`."
    ) from e
from structlog.stdlib import BoundLogger

from ._manager import PydanticSchemaManager

__all__ = [
    "SchemaManagerSettings",
    "SchemaRegistryClientParams",
]


class SchemaRegistryClientParams(TypedDict):
    """Kwargs used to construct an AsyncSchemaRegistryClient."""

    url: str
    """URL of a a Confluent-compatible schema registry."""


class SchemaManagerSettings(BaseSettings):
    """Settings for constructing a `~safir.kafka.PydanticSchemaManager`."""

    model_config = SettingsConfigDict(
        case_sensitive=False,
        env_prefix="SCHEMA_MANAGER_",
        extra="forbid",
        populate_by_name=True,
    )

    registry_url: AnyUrl = Field(
        title="Schema registry URL",
        description="URL of a a Confluent-compatible schema registry",
        validation_alias=AliasChoices(
            "registryUrl", "SCHEMA_MANAGER_REGISTRY_URL"
        ),
    )

    suffix: str = Field(
        default="",
        title="Suffix",
        description=(
            "A suffix that is added to the schema name (and thus the subject"
            " name). The suffix creates alternate subjects in the Schema"
            " Registry so schemas registered during testing and staging don't"
            " affect the compatibility continuity of a production subject. For"
            " production, it's best to not set a suffix."
        ),
        examples=["_dev1"],
    )

    def to_registry_params(self) -> SchemaRegistryClientParams:
        """Make a dict of params to construct an AsyncSchemaRegistryClient."""
        return {"url": str(self.registry_url)}

    def make_manager(
        self, logger: BoundLogger | None = None
    ) -> PydanticSchemaManager:
        """Construct a PydanticSchemaManager from the fields of this model."""
        registry = AsyncSchemaRegistryClient(**self.to_registry_params())
        return PydanticSchemaManager(
            registry=registry, suffix=self.suffix, logger=logger
        )



"""Tests for datetime utility functions."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta, timezone

import pytest
from pydantic import BaseModel

from safir.datetime import (
    current_datetime,
    format_datetime_for_logging,
    isodatetime,
    parse_isodatetime,
    parse_timedelta,
)


def test_current_datetime() -> None:
    time = current_datetime()
    assert time.microsecond == 0
    assert time.tzinfo == UTC
    now = datetime.now(tz=UTC)
    assert now - timedelta(seconds=2) <= time <= now

    time = current_datetime(microseconds=True)
    if not time.microsecond:
        time = current_datetime(microseconds=True)
    assert time.microsecond != 0
    assert time.tzinfo == UTC
    now = datetime.now(tz=UTC)
    assert now - timedelta(seconds=2) <= time <= now


def test_isodatetime() -> None:
    time = datetime.fromisoformat("2022-09-16T12:03:45+00:00")
    assert isodatetime(time) == "2022-09-16T12:03:45Z"

    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        isodatetime(datetime.fromisoformat("2022-09-16T12:03:45+02:00"))

    # Pydantic's JSON decoder uses a TzInfo data structure instead of
    # datetime.timezone.utc. Make sure that's still recognized as UTC.
    class Test(BaseModel):
        time: datetime

    json_model = Test(time=time).model_dump_json()
    model = Test.model_validate_json(json_model)
    assert isodatetime(model.time) == "2022-09-16T12:03:45Z"


def test_parse_isodatetime() -> None:
    time = parse_isodatetime("2022-09-16T12:03:45Z")
    assert time == datetime(2022, 9, 16, 12, 3, 45, tzinfo=UTC)
    now = current_datetime()
    assert parse_isodatetime(isodatetime(now)) == now
    time = parse_isodatetime("2022-09-16T12:03:45")
    assert time == datetime(2022, 9, 16, 12, 3, 45, tzinfo=UTC)
    time = parse_isodatetime("2022-09-16")
    assert time == datetime(2022, 9, 16, 0, 0, 0, tzinfo=UTC)

    with pytest.raises(ValueError, match="does not match IVOA format"):
        parse_isodatetime("2022-09-16T12:03:45+00:00")


def test_format_datetime_for_logging() -> None:
    time = datetime.fromisoformat("2022-09-16T12:03:45+00:00")
    assert format_datetime_for_logging(time) == "2022-09-16 12:03:45"

    # Test with milliseconds, allowing for getting extremely unlucky and
    # having no microseconds. Getting unlucky twice seems impossible, so we'll
    # fail in that case rather than loop.
    now = datetime.now(tz=UTC)
    if not now.microsecond:
        now = datetime.now(tz=UTC)
    milliseconds = int(now.microsecond / 1000)
    expected = now.strftime("%Y-%m-%d %H:%M:%S") + f".{milliseconds:03n}"
    assert format_datetime_for_logging(now) == expected

    time = datetime.now(tz=timezone(timedelta(hours=1)))
    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        format_datetime_for_logging(time)

    # Pydantic's JSON decoder uses a TzInfo data structure instead of
    # datetime.timezone.utc. Make sure that's still recognized as UTC.
    class Test(BaseModel):
        time: datetime

    json_model = Test(time=now).model_dump_json()
    model = Test.model_validate_json(json_model)
    assert format_datetime_for_logging(model.time) == expected


def test_parse_timdelta() -> None:
    assert parse_timedelta("8d") == timedelta(days=8)
    assert parse_timedelta("4h 3minutes\n") == timedelta(hours=4, minutes=3)
    assert parse_timedelta("\n 5w4d") == timedelta(weeks=5, days=4)
    assert parse_timedelta(
        "2weeks 2days 2hours 2minutes 2seconds"
    ) == timedelta(weeks=2, days=2, hours=2, minutes=2, seconds=2)
    assert parse_timedelta("1week 1day 1hour 1minute 1second") == timedelta(
        weeks=1, days=1, hours=1, minutes=1, seconds=1
    )
    assert parse_timedelta("4hr 5mins 6secs") == timedelta(
        hours=4, minutes=5, seconds=6
    )
    assert parse_timedelta("17min 65sec") == timedelta(minutes=17, seconds=65)

    with pytest.raises(ValueError, match="Could not parse"):
        parse_timedelta("3s4m")
    with pytest.raises(ValueError, match="Could not parse"):
        parse_timedelta("4weeks 36ms")



"""Functions to format dates and times."""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import overload

__all__ = [
    "format_datetime_for_logging",
    "isodatetime",
]


@overload
def format_datetime_for_logging(timestamp: datetime) -> str: ...


@overload
def format_datetime_for_logging(timestamp: None) -> None: ...


def format_datetime_for_logging(timestamp: datetime | None) -> str | None:
    """Format a datetime for logging and human readabilty.

    Parameters
    ----------
    timestamp
        Object to format. Must be in UTC or timezone-naive (in which case it's
        assumed to be in UTC).

    Returns
    -------
    str or None
        The datetime in format ``YYYY-MM-DD HH:MM:SS[.sss]`` with milliseconds
        added if and only if the microseconds portion of ``timestamp`` is not
        0. There will be no ``T`` separator or time zone information.

    Raises
    ------
    ValueError
        Raised if the argument is in a time zone other than UTC.
    """
    if timestamp:
        if timestamp.utcoffset() != timedelta(seconds=0):
            raise ValueError(f"datetime {timestamp} not in UTC")
        if timestamp.microsecond:
            result = timestamp.isoformat(sep=" ", timespec="milliseconds")
        else:
            result = timestamp.isoformat(sep=" ", timespec="seconds")
        return result.split("+")[0]
    else:
        return None


def isodatetime(timestamp: datetime) -> str:
    """Format a timestamp in UTC in a standard ISO date format.

    Parameters
    ----------
    timestamp
        Date and time to format.

    Returns
    -------
    str
        Date and time formatted as an ISO 8601 date and time using ``Z`` as
        the time zone.  This format is compatible with both Kubernetes and the
        IVOA UWS standard.

    Raises
    ------
    ValueError
        The provided timestamp was not in UTC.
    """
    if timestamp.utcoffset() != timedelta(seconds=0):
        raise ValueError(f"datetime {timestamp} not in UTC")
    return timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")



"""Fixtures for captured Sentry information and transport.

Taken from the `Sentry Python SDK's unit tests
<https://github.com/getsentry/sentry-python/blob/c6a89d64db965fe0ece6de10df38ab936af8f5e4/tests/conftest.py>`__.
"""

# Copyright (c) 2018 Functional Software, Inc. dba Sentry
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import contextlib
from collections.abc import Callable, Generator
from typing import Any

import pytest
import sentry_sdk
from sentry_sdk.envelope import Envelope

from ._mocks import Attachment, Captured, TestTransport

__all__ = [
    "capture_events_fixture",
    "sentry_init_fixture",
]


@contextlib.contextmanager
def sentry_init_fixture() -> Generator[Callable[..., None]]:
    """Return an init function that injects a no-op transport."""

    def inner(*a: Any, **kw: Any) -> None:
        kw.setdefault("transport", TestTransport())
        client = sentry_sdk.Client(*a, **kw)
        sentry_sdk.get_global_scope().set_client(client)

    old_client = sentry_sdk.get_global_scope().client
    try:
        sentry_sdk.get_current_scope().set_client(None)
        yield inner
    finally:
        sentry_sdk.get_global_scope().set_client(old_client)


def capture_events_fixture(
    monkeypatch: pytest.MonkeyPatch,
) -> Callable[[], Captured]:
    """Return a function that returns a container with items sent to Sentry."""

    def inner() -> Captured:
        test_client = sentry_sdk.get_client()
        if test_client.transport is None:
            msg = "Error patching Sentry transport: client.transport is None"
            raise RuntimeError(msg)
        old_capture_envelope = test_client.transport.capture_envelope

        captured = Captured(errors=[], transactions=[], attachments=[])

        def append(envelope: Envelope) -> None:
            for item in envelope:
                match item.headers.get("type"):
                    case "event":
                        if item.payload.json is None:
                            raise ValueError(
                                "Sentry event unexpectedly missing json"
                                " payload"
                            )
                        captured.errors.append(item.payload.json)
                    case "transaction":
                        if item.payload.json is None:
                            raise ValueError(
                                "Sentry transaction unexpectedly missing json"
                                " payload"
                            )
                        captured.transactions.append(item.payload.json)
                    case "attachment":
                        captured.attachments.append(
                            Attachment(
                                filename=item.headers["filename"],
                                bytes=item.payload.get_bytes(),
                                content_type=item.headers["content_type"],
                            )
                        )
                    case _:
                        pass
            return old_capture_envelope(envelope)

        monkeypatch.setattr(test_client.transport, "capture_envelope", append)

        return captured

    return inner



"""V1 schema

Revision ID: a6e49b57e3c9
Revises:
Create Date: 2024-09-10 23:04:48.696687+00:00

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "a6e49b57e3c9"
down_revision: str | None = None
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column("username", sa.String(), nullable=False),
        sa.Column("name", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("username"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("users")
    # ### end Alembic commands ###



"""Alembic support using an async database backend."""

from __future__ import annotations

import asyncio
from contextlib import suppress
from pathlib import Path

try:
    import alembic
    from alembic import context
    from alembic.config import Config
    from alembic.runtime.migration import MigrationContext
    from alembic.script import ScriptDirectory
    from sqlalchemy import MetaData, text
    from sqlalchemy.engine import Connection
    from sqlalchemy.exc import ProgrammingError
    from sqlalchemy.ext.asyncio import AsyncEngine
except ImportError as e:
    raise ImportError(
        "The safir.database module requires the db extra. "
        "Install it with `pip install safir[db]`."
    ) from e
from pydantic import SecretStr
from pydantic_core import Url
from structlog.stdlib import BoundLogger

from ._connection import build_database_url, create_database_engine

__all__ = [
    "AlembicConfigError",
    "is_database_current",
    "run_migrations_offline",
    "run_migrations_online",
    "stamp_database",
    "stamp_database_async",
    "unstamp_database",
]


class AlembicConfigError(Exception):
    """The Alembic configuration was missing or invalid."""


async def is_database_current(
    engine: AsyncEngine,
    logger: BoundLogger | None = None,
    config_path: Path = Path("alembic.ini"),
) -> bool:
    """Check whether the database schema is at the current version.

    Every entry point to the application other than ones dedicated to updating
    the schema should normally call this function after creating the database
    engine and abort if it returns `False`.

    Parameters
    ----------
    engine
        Database engine to use.
    logger
        Logger to which the schema mismatch will be reported if the schema is
        out of date.
    config_path
        Path to the Alembic configuration.

    Returns
    -------
    bool
        `True` if Alembic reports the database schema is current, `False`
        otherwise.
    """

    def get_current_heads(connection: Connection) -> set[str]:
        context = MigrationContext.configure(connection)
        return set(context.get_current_heads())

    async with engine.begin() as connection:
        current = await connection.run_sync(get_current_heads)
    alembic_config = Config(str(config_path))
    alembic_scripts = ScriptDirectory.from_config(alembic_config)
    expected = set(alembic_scripts.get_heads())
    if current != expected:
        if logger:
            logger.error(
                f"Schema mismatch: {current} != {expected}",
                current_schema=list(current),
                expected_schema=list(expected),
            )
        return False
    else:
        return True


def run_migrations_offline(metadata: MetaData, url: str | Url) -> None:
    """Run Alembic migrations in offline mode.

    This function may only be called from the Alembic :file:`env.py` file.

    This configures the context with just a URL and not an engine. By skipping
    the engine creation, the DBAPI does not have to be available. The
    migrations are sent to the script output.

    Parameters
    ----------
    metadata
        Schema metadata object for the current schema.
    url
        Database connection URL.

    Examples
    --------
    Normally this is called from :file:`alembic/env.py` with code similar to
    the following:

    .. code-block:: python

       from alembic import context
       from safir.database import run_migrations_offline

       from example.config import config
       from example.schema import Base


       if context.is_offline_mode():
           run_migrations_offline(Base.metadata, config.database_url)
    """
    url = build_database_url(url, None, is_async=False)
    context.configure(
        url=url,
        target_metadata=metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online(
    metadata: MetaData,
    url: str | Url,
    password: str | SecretStr | None,
) -> None:
    """Run Alembic migrations online using an async backend.

    This function may only be called from the Alembic :file:`env.py` file.

    Parameters
    ----------
    metadata
        Schema metadata object for the current schema.
    url
        Database connection URL, not including the password.
    password
        Database connection password.

    Examples
    --------
    Normally this is called from :file:`alembic/env.py` with code similar to
    the following:

    .. code-block:: python

       from alembic import context
       from safir.database import run_migrations_offline

       from example.config import config
       from example.schema import Base


       if not context.is_offline_mode():
           run_migrations_online(
               Base.metadata, config.database_url, config.database_password
           )
    """

    def do_migrations(connection: Connection) -> None:
        context.configure(connection=connection, target_metadata=metadata)
        with context.begin_transaction():
            context.run_migrations()

    async def run_async_migrations() -> None:
        engine = create_database_engine(url, password)
        async with engine.connect() as connection:
            await connection.run_sync(do_migrations)
        await engine.dispose()

    asyncio.run(run_async_migrations())


def stamp_database(config_path: Path) -> None:
    """Mark the database as updated to the head of the given Alembic config.

    Add the necessary data to the database to mark it as having the current
    schema, without performing any migrations. It is intended to be called
    immediately after a fresh database initialization with the current schema.

    Parameters
    ----------
    config_path
        Path to the Alembic configuration.
    """
    alembic.command.stamp(Config(str(config_path)), "head")


async def stamp_database_async(
    engine: AsyncEngine, config_path: Path = Path("alembic.ini")
) -> None:
    """Mark the database as updated to the head of the given Alembic config.

    Add the necessary data to the database to mark it as having the current
    schema, without performing any migrations. It is intended to be called
    immediately after a fresh database initialization with the current schema.

    The Alembic configuration must be :file:`alembic.ini` in the current
    directory, and contain correct pointers to the database migrations.

    If running outside of an event loop, use `stamp_database` instead.

    Parameters
    ----------
    engine
        Database engine to use.
    config_path
        Path to the Alembic configuration.

    Raises
    ------
    AlembicConfigError
        Raised if no migration heads could be found in the Alembic
        configuration.
    """
    alembic_config = Config(str(config_path))
    alembic_scripts = ScriptDirectory.from_config(alembic_config)
    current_head = alembic_scripts.get_current_head()
    if not current_head:
        raise AlembicConfigError("Unable to find head of Alembic migrations")

    def set_version(connection: Connection) -> None:
        context = MigrationContext.configure(connection)
        context.stamp(alembic_scripts, current_head)

    await unstamp_database(engine)
    async with engine.begin() as connection:
        await connection.run_sync(set_version)


async def unstamp_database(engine: AsyncEngine) -> None:
    """Clear the Alembic version from the database.

    This is done before stamping the database, or in test fixtures that are
    restoring a test database to an uninitialized state.

    Parameters
    ----------
    engine
        Engine to use for SQL calls.
    """
    async with engine.begin() as conn:
        with suppress(ProgrammingError):
            await conn.execute(text("DROP TABLE alembic_version"))



"""An arq_ client with a mock for testing."""

from ._config import WorkerSettings, build_arq_redis_settings
from ._exceptions import (
    ArqJobError,
    JobNotFound,
    JobNotQueued,
    JobResultUnavailable,
)
from ._models import ArqMode, JobMetadata, JobResult
from ._queue import ArqQueue, MockArqQueue, RedisArqQueue

__all__ = [
    "ArqJobError",
    "ArqMode",
    "ArqQueue",
    "JobMetadata",
    "JobNotFound",
    "JobNotQueued",
    "JobResult",
    "JobResultUnavailable",
    "MockArqQueue",
    "RedisArqQueue",
    "WorkerSettings",
    "build_arq_redis_settings",
]



"""datetime management for databases."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta
from typing import overload

__all__ = [
    "datetime_from_db",
    "datetime_to_db",
]


@overload
def datetime_from_db(time: datetime) -> datetime: ...


@overload
def datetime_from_db(time: None) -> None: ...


def datetime_from_db(time: datetime | None) -> datetime | None:
    """Add the UTC time zone to a naive datetime from the database.

    Parameters
    ----------
    time
        The naive datetime from the database, or `None`.

    Returns
    -------
    datetime.datetime or None
        `None` if the input was none, otherwise a timezone-aware version of
        the same `~datetime.datetime` in the UTC timezone.
    """
    if not time:
        return None
    if time.tzinfo not in (None, UTC):
        raise ValueError(f"datetime {time} not in UTC")
    return time.replace(tzinfo=UTC)


@overload
def datetime_to_db(time: datetime) -> datetime: ...


@overload
def datetime_to_db(time: None) -> None: ...


def datetime_to_db(time: datetime | None) -> datetime | None:
    """Strip time zone for storing a datetime in the database.

    Parameters
    ----------
    time
        The timezone-aware `~datetime.datetime` in the UTC time zone, or
        `None`.

    Returns
    -------
    datetime.datetime or None
        `None` if the input was `None`, otherwise the same
        `~datetime.datetime` but timezone-naive and thus suitable for storing
        in a SQL database.
    """
    if not time:
        return None
    if time.utcoffset() != timedelta(seconds=0):
        raise ValueError(f"datetime {time} not in UTC")
    return time.replace(tzinfo=None)



"""Manage an async database session."""

from collections.abc import AsyncGenerator
from typing import Any

from pydantic import SecretStr
from pydantic_core import Url
from sqlalchemy.ext.asyncio import AsyncEngine, async_scoped_session

from safir.database import create_async_session, create_database_engine

__all__ = ["DatabaseSessionDependency", "db_session_dependency"]


class DatabaseSessionDependency:
    """Manages an async per-request SQLAlchemy session.

    Notes
    -----
    Creation of the database session factory has to be deferred until the
    configuration has been loaded, which in turn is deferred until app
    startup.

    In the app startup hook, run:

    .. code-block:: python

       await db_session_dependency.initialize(database_url)

    In the app shutdown hook, run:

    .. code-block:: python

       await db_session_dependency.aclose()

    An isolation level may optionally be configured when calling `initialize`.
    By default, a transaction is opened for every request and committed at the
    end of that request.  This can be configured when calling `initialize`.
    """

    def __init__(self) -> None:
        self._engine: AsyncEngine | None = None
        self._session: async_scoped_session | None = None

    async def __call__(self) -> AsyncGenerator[async_scoped_session]:
        """Return the database session manager.

        Returns
        -------
        sqlalchemy.ext.asyncio.AsyncSession
            The newly-created session.
        """
        if not self._session:
            raise RuntimeError("db_session_dependency not initialized")
        try:
            yield self._session
        finally:
            # Following the recommendations in the SQLAlchemy documentation,
            # each session is scoped to a single web request. However, this
            # all uses the same async_scoped_session object, so should share
            # an underlying engine and connection pool.
            await self._session.remove()

    async def aclose(self) -> None:
        """Shut down the database engine."""
        if self._engine:
            await self._engine.dispose()
            self._engine = None

    async def initialize(
        self,
        url: str | Url,
        password: str | SecretStr | None,
        *,
        connect_args: dict[str, Any] | None = None,
        isolation_level: str | None = None,
        max_overflow: int | None = None,
        pool_size: int | None = None,
        pool_timeout: float | None = None,
    ) -> None:
        """Initialize the session dependency.

        Parameters
        ----------
        url
            Database connection URL, not including the password.
        password
            Database connection password.
        connect_args
            Additional connection arguments to pass directly to the underlying
            database driver.
        isolation_level
            If specified, sets a non-default isolation level for the database
            engine.
        max_overflow
            Maximum number of connections over the pool size for surge
            traffic.
        pool_size
            Connection pool size.
        pool_timeout
            How long to wait for a connection from the connection pool before
            giving up.
        """
        if self._engine:
            await self._engine.dispose()
        kwargs: dict[str, Any] = {}
        if connect_args:
            kwargs["connect_args"] = connect_args
        if isolation_level:
            kwargs["isolation_level"] = isolation_level
        if max_overflow is not None:
            kwargs["max_overflow"] = max_overflow
        if pool_size is not None:
            kwargs["pool_size"] = pool_size
        if pool_timeout is not None:
            kwargs["pool_timeout"] = pool_timeout
        self._engine = create_database_engine(url, password, **kwargs)
        self._session = await create_async_session(self._engine)


db_session_dependency = DatabaseSessionDependency()
"""The dependency that will return the async session proxy."""



"""V2 schema

Revision ID: e843cec7e694
Revises: a6e49b57e3c9
Create Date: 2024-09-10 23:16:17.755569+00:00

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "e843cec7e694"
down_revision: str | None = "a6e49b57e3c9"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("users", sa.Column("email", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("users", "email")
    # ### end Alembic commands ###



"""Models used by the arq_ client."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from typing import Any, Self, override

from arq.jobs import Job, JobStatus

from ._exceptions import JobNotFound, JobResultUnavailable

__all__ = [
    "ArqMode",
    "JobMetadata",
    "JobResult",
]


class ArqMode(StrEnum):
    """Mode configuration for the Arq queue."""

    production = "production"
    """Normal usage of arq, with a Redis broker."""

    test = "test"
    """Use the MockArqQueue to test an API service without standing up a
    full distributed worker queue.
    """


@dataclass
class JobMetadata:
    """Information about a queued job."""

    id: str
    """The `~arq.jobs.Job` identifier."""

    name: str
    """The task name."""

    args: tuple[Any, ...]
    """The positional arguments to the task function."""

    kwargs: dict[str, Any]
    """The keyword arguments to the task function."""

    enqueue_time: datetime
    """Time when the job was added to the queue."""

    status: JobStatus
    """Status of the job.

    States are defined by the `arq.jobs.JobStatus` enumeration:

    - ``deferred`` (in queue, but waiting a predetermined time to become
      ready to run)
    - ``queued`` (queued to run)
    - ``in_progress`` (actively being run by a worker)
    - ``complete`` (result is available)
    - ``not_found`` (the job cannot be found)
    """

    queue_name: str
    """Name of the queue this job belongs to."""

    @classmethod
    async def from_job(cls, job: Job) -> Self:
        """Initialize JobMetadata from an arq Job.

        Raises
        ------
        JobNotFound
            Raised if the job is not found
        """
        job_info = await job.info()
        if job_info is None:
            raise JobNotFound(job.job_id)

        job_status = await job.status()
        if job_status == JobStatus.not_found:
            raise JobNotFound(job.job_id)

        return cls(
            id=job.job_id,
            name=job_info.function,
            args=job_info.args,
            kwargs=job_info.kwargs,
            enqueue_time=job_info.enqueue_time,
            status=job_status,
            # private attribute of Job; not available in JobDef
            # queue_name is available in JobResult
            queue_name=job._queue_name,  # noqa: SLF001
        )


@dataclass
class JobResult(JobMetadata):
    """The full result of a job, as well as its metadata."""

    start_time: datetime
    """Time when the job started."""

    finish_time: datetime
    """Time when the job finished."""

    success: bool
    """`True` if the job returned without an exception, `False` if an
    exception was raised.
    """

    result: Any
    """The job's result."""

    @override
    @classmethod
    async def from_job(cls, job: Job) -> Self:
        """Initialize the `JobResult` from an arq `~arq.jobs.Job`.

        Raises
        ------
        JobNotFound
            Raised if the job is not found
        JobResultUnavailable
            Raised if the job result is not available.
        """
        job_info = await job.info()
        if job_info is None:
            raise JobNotFound(job.job_id)

        job_status = await job.status()
        if job_status == JobStatus.not_found:
            raise JobNotFound(job.job_id)

        # Result may be none if the job isn't finished
        result_info = await job.result_info()
        if result_info is None:
            raise JobResultUnavailable(job.job_id)

        return cls(
            id=job.job_id,
            name=job_info.function,
            args=job_info.args,
            kwargs=job_info.kwargs,
            enqueue_time=job_info.enqueue_time,
            start_time=result_info.start_time,
            finish_time=result_info.finish_time,
            success=result_info.success,
            status=job_status,
            queue_name=result_info.queue_name,
            result=result_info.result,
        )



"""Sentry helpers."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta

import sentry_sdk
from sentry_sdk.attachments import Attachment
from sentry_sdk.tracing import Span
from sentry_sdk.types import Event, Hint

from safir.slack.webhook import SlackIgnoredException, SlackWebhookClient

from ..slack.blockkit import SlackException

__all__ = [
    "before_send_handler",
    "duration",
    "fingerprint_env_handler",
    "report_exception",
    "sentry_exception_handler",
]


def duration(span: Span) -> timedelta:
    """Return the time spent in a span (to the present if not finished)."""
    if span.timestamp is None:
        timestamp = datetime.now(tz=UTC)
    else:
        timestamp = span.timestamp
    return timestamp - span.start_timestamp


def fingerprint_env_handler(event: Event, _: Hint) -> Event:
    """Add the environment to the event fingerprint.

    Without doing this, Sentry groups events from all environments into the
    same issue, and alerts that notify on new issues won't notify on a prod
    event if an issue has already been created from an event from another env
    :(

    https://github.com/getsentry/sentry/issues/64354
    """
    env = event.get("environment")
    if env is None:
        env = "no environment"
    fingerprint = event.get("fingerprint", [])
    event["fingerprint"] = [
        "{{ default }}",
        *fingerprint,
        env,
    ]
    return event


def sentry_exception_handler(event: Event, hint: Hint) -> Event:
    """Add tags and context from `~safir.slack.blockkit.SlackException`.

    The Sentry event data model is:
    https://develop.sentry.dev/sdk/data-model/event-payloads/
    """
    if exc_info := hint.get("exc_info"):
        exc = exc_info[1]
        if isinstance(exc, SlackException):
            info = exc.to_sentry()
            event["tags"] = event.setdefault("tags", {})
            event["tags"].update(info.tags)
            event["contexts"] = event.setdefault("contexts", {})
            event["contexts"].update(info.contexts)

            if info.username:
                event["user"] = event.setdefault("user", {})
                event["user"]["username"] = info.username

            for key, content in info.attachments.items():
                attachments = hint.setdefault("attachments", [])
                attachments.append(
                    Attachment(filename=key, bytes=content.encode())
                )

    return event


def before_send_handler(event: Event, hint: Hint) -> Event:
    """Add the env to the fingerprint, and enrich from
    `~safir.slack.blockkit.SlackException`.
    """
    event = fingerprint_env_handler(event, hint)
    return sentry_exception_handler(event, hint)


async def report_exception(
    exc: Exception, slack_client: SlackWebhookClient | None
) -> None:
    """Notify Sentry and Slack about an exception.

    Use this when you want to be notified about an exception, but stop
    propagating it. If you want to keep propagating the exception and still
    support both Slack and Sentry, catch the exception, use
    `~safir.slack.webhook.SlackWebhookClient.post_exception`, re-raise the
    exception, and let the Sentry-configured uncaught exception handler notify
    Sentry if the exception makes it that far.

    If Sentry has not been initialized, then no events will be sent to Sentry.
    If ``slack_client`` is `None`, then no notifications will be sent to
    Slack.

    If the exception is an instance of
    `~safir.slack.webhook.SlackIgnoredException`, no notifications will be
    sent to either Sentry or Slack. If the exception is an instance of
    `~safir.slack.blockkit.SlackException`, the Slack message will be generated
    from its ``to_slack`` method.

    Parameters
    ----------
    exc
        The Exception to send to the notification service.

    slack_client
        A client to use to send a Slack notification about the exception. If \
        this is `None`, then no Slack notification will be sent.
    """
    if isinstance(exc, SlackIgnoredException):
        return

    sentry_sdk.capture_exception(exc)
    if slack_client:
        await slack_client.post_exception(exc)
